from __future__ import annotations

from typing import ClassVar, Literal

import equinox as eqx
from jax import lax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Bool, Float, Int, Key

from lerax.render import AbstractRenderer, Color, PygameRenderer, Transform
from lerax.space import Box, Discrete

from .base_env import AbstractEnv


class MountainCar(AbstractEnv[Int[Array, ""], Float[Array, "2"]]):
    name: ClassVar[str] = "MountainCar"

    state_index: eqx.nn.StateIndex[Float[Array, "2"]]
    action_space: Discrete
    observation_space: Box

    min_position: float
    max_position: float
    max_speed: float
    goal_position: float
    goal_velocity: float

    force: float
    gravity: float
    low: Float[Array, "2"]
    high: Float[Array, "2"]

    renderer: AbstractRenderer | None

    def __init__(
        self,
        goal_velocity: float = 0,
        renderer: AbstractRenderer | Literal["auto"] | None = None,
    ):
        self.min_position = -1.2
        self.max_position = 0.6
        self.max_speed = 0.07
        self.goal_position = 0.5
        self.goal_velocity = goal_velocity

        self.force = 0.001
        self.gravity = 0.0025

        self.low = jnp.array([self.min_position, -self.max_speed])
        self.high = jnp.array([self.max_position, self.max_speed])

        self.action_space = Discrete(3)
        self.observation_space = Box(self.low, self.high)

        self.state_index = eqx.nn.StateIndex(jnp.zeros(2))

        if renderer == "auto":
            self.renderer = self.default_renderer()
        else:
            self.renderer = renderer

    def reset(
        self, state: eqx.nn.State, *, key: Key, low: float = -0.6, high: float = -0.4
    ) -> tuple[eqx.nn.State, Float[Array, "2"], dict]:
        position = jr.uniform(key, minval=low, maxval=high)
        velocity = 0.0

        state_vals = jnp.asarray([position, velocity])
        state = state.set(self.state_index, state_vals)

        return state, state_vals, {}

    def step(self, state: eqx.nn.State, action: Int[Array, ""], *, key: Key) -> tuple[
        eqx.nn.State,
        Float[Array, "2"],
        Float[Array, ""],
        Bool[Array, ""],
        Bool[Array, ""],
        dict,
    ]:
        position, velocity = state.get(self.state_index)

        velocity += (action - 1) * self.force + jnp.cos(3 * position) * (-self.gravity)
        velocity = jnp.clip(velocity, -self.max_speed, self.max_speed)

        position += velocity
        position = jnp.clip(position, self.min_position, self.max_position)

        velocity = lax.cond(
            jnp.logical_and(position == self.min_position, velocity < 0),
            lambda: jnp.array(0.0),
            lambda: velocity,
        )

        terminated = jnp.logical_and(
            position >= self.goal_position, velocity >= self.goal_velocity
        )
        reward = jnp.array(-1.0)

        state_vals = jnp.asarray([position, velocity])
        state = state.set(self.state_index, state_vals)

        return state, state_vals, reward, terminated, jnp.array(False), {}

    def render(self, state: eqx.nn.State):
        x, _ = state.get(self.state_index)

        assert self.renderer is not None, "Renderer is not initialized."
        self.renderer.clear()

        # Track
        xs = jnp.linspace(self.min_position - 0.5, self.max_position + 0.1, 64)
        ys = jnp.sin(3 * xs) * 0.45
        track_points = jnp.stack([xs, ys], axis=1)
        track_color = Color(0.0, 0.0, 0.0)
        self.renderer.draw_polyline(track_points, color=track_color)

        # Flag
        flag_h = 0.2
        flag_start = jnp.array(
            [self.goal_position, jnp.sin(3 * self.goal_position) * 0.45]
        )
        flag_end = flag_start + jnp.array([0.0, flag_h])

        flag_pole_color = Color(0.0, 0.0, 0.0)
        flag_color = Color(0.86, 0.24, 0.24)
        self.renderer.draw_line(
            flag_start, flag_end, color=flag_pole_color, width=0.005
        )
        flag_points = jnp.array(
            [
                flag_end,
                flag_end + jnp.array([0.1, -0.03]),
                flag_end + jnp.array([0.0, -0.06]),
            ]
        )
        self.renderer.draw_polygon(flag_points, color=flag_color)

        # Car
        car_h, car_w, wheel_r, clearance = 0.04, 0.1, 0.02, 0.025
        angle = jnp.arctan2(jnp.cos(3 * x) * 0.45 * 3, 1.0)
        rot = jnp.array(
            [[jnp.cos(angle), -jnp.sin(angle)], [jnp.sin(angle), jnp.cos(angle)]]
        )
        ## Body
        car_col = Color(0.0, 0.0, 0.0)
        clearance_vec = rot @ jnp.array([0.0, clearance + car_h / 2])
        car_center = jnp.array([x, jnp.sin(3 * x) * 0.45]) + clearance_vec

        car_corners = jnp.array(
            [
                [-car_w / 2, -car_h / 2],
                [car_w / 2, -car_h / 2],
                [car_w / 2, car_h / 2],
                [-car_w / 2, car_h / 2],
            ]
        )
        car_corners = (rot @ car_corners.T).T + car_center
        self.renderer.draw_polygon(car_corners, color=car_col)
        ## Wheels
        wheel_col = Color(0.3, 0.3, 0.3)
        wheel_clearance = rot @ jnp.array([0.0, wheel_r])
        wheel_centers = jnp.array(
            [
                [-car_w / 4, 0.0],
                [car_w / 4, 0.0],
            ]
        )
        wheel_centers = (
            (rot @ wheel_centers.T).T
            + jnp.array([x, jnp.sin(3 * x) * 0.45])
            + wheel_clearance
        )
        self.renderer.draw_circle(wheel_centers[0], radius=wheel_r, color=wheel_col)
        self.renderer.draw_circle(wheel_centers[1], radius=wheel_r, color=wheel_col)

        self.renderer.render()

    def default_renderer(self) -> AbstractRenderer:
        width, height = 800, 450
        transform = Transform(
            scale=350.0,
            offset=jnp.array([width * 0.7, height * 0.4]),
            width=width,
            height=height,
            y_up=True,
        )
        return PygameRenderer(
            width=width,
            height=height,
            background_color=Color(1.0, 1.0, 1.0),
            transform=transform,
        )

    def close(self):
        pass



from __future__ import annotations

from typing import Any

from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key

from .base_space import AbstractSpace
from .utils import try_cast


class MultiBinary(AbstractSpace[Bool[Array, " n"]]):
    """A space of binary values."""

    n: int

    def __init__(self, n: int):
        assert n > 0, "n must be positive"
        self.n = n

    @property
    def shape(self) -> tuple[int, ...]:
        return (self.n,)

    def canonical(self) -> Bool[Array, " n"]:
        return jnp.zeros(self.shape, dtype=bool)

    def sample(self, key: Key) -> Bool[Array, " n"]:
        return jr.bernoulli(key, shape=self.shape)

    def contains(self, x: Any) -> Bool[ArrayLike, ""]:
        x = try_cast(x)
        if x is None:
            return False

        if x.shape != self.shape:
            return False

        return jnp.all((x == 0) | (x == 1), axis=0)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MultiBinary):
            return False

        return bool(self.n == other.n)

    def __repr__(self) -> str:
        return f"MultiBinary({self.n})"

    def __hash__(self) -> int:
        return hash(self.n)

    def flatten_sample(self, sample: Bool[ArrayLike, " n"]) -> Float[Array, " size"]:
        return jnp.asarray(sample, dtype=float).ravel()

    @property
    def flat_size(self) -> Int[ArrayLike, ""]:
        return jnp.array(self.n, dtype=int)



from __future__ import annotations

import equinox as eqx
import jax
import optax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Float, Key, Scalar

from lerax.buffer import RolloutBuffer
from lerax.env import AbstractEnvLike
from lerax.policy import AbstractActorCriticPolicy
from lerax.utils import filter_scan

from .on_policy import AbstractOnPolicyAlgorithm


class PPOStats(eqx.Module):
    approx_kl: Float[Array, ""]
    total_loss: Float[Array, ""]
    policy_loss: Float[Array, ""]
    value_loss: Float[Array, ""]
    entropy_loss: Float[Array, ""]
    state_magnitude_loss: Float[Array, ""]


class PPO[ActType, ObsType](AbstractOnPolicyAlgorithm[ActType, ObsType]):
    """Proximal Policy Optimization (PPO) algorithm."""

    env: AbstractEnvLike[ActType, ObsType]
    policy: AbstractActorCriticPolicy[Float, ActType, ObsType]
    optimizer: eqx.AbstractVar[optax.GradientTransformation]
    opt_state_index: eqx.AbstractVar[eqx.nn.StateIndex[optax.OptState]]

    clipper: optax.GradientTransformation
    clipper_index: eqx.nn.StateIndex

    gae_lambda: float = eqx.field(static=True)
    gamma: float = eqx.field(static=True)
    num_steps: int = eqx.field(static=True)
    batch_size: int = eqx.field(static=True)

    num_epochs: int = eqx.field(static=True)
    num_mini_batches: int = eqx.field(static=True)

    normalize_advantages: bool = eqx.field(static=True)
    clip_coefficient: float = eqx.field(static=True)
    clip_value_loss: bool = eqx.field(static=True)
    entropy_loss_coefficient: float = eqx.field(static=True)
    value_loss_coefficient: float = eqx.field(static=True)
    state_magnitude_coefficient: float = eqx.field(static=True)
    max_grad_norm: float = eqx.field(static=True)

    def __init__(
        self,
        env: AbstractEnvLike[ActType, ObsType],
        policy: AbstractActorCriticPolicy[Float, ActType, ObsType],
        *,
        num_steps: int = 2048,
        num_epochs: int = 16,
        num_batches: int = 32,
        gae_lambda: float = 0.95,
        gamma: float = 0.99,
        clip_coefficient: float = 0.2,
        clip_value_loss: bool = False,
        entropy_loss_coefficient: float = 0.0,
        value_loss_coefficient: float = 0.5,
        state_magnitude_coefficient: float = 0.0,
        max_grad_norm: float = 0.5,
        normalize_advantages: bool = True,
    ):
        self.env = env
        self.policy = policy

        self.num_steps = int(num_steps)
        self.num_epochs = int(num_epochs)
        self.num_mini_batches = int(num_batches)
        self.batch_size = self.num_steps // self.num_mini_batches

        self.gae_lambda = float(gae_lambda)
        self.gamma = float(gamma)

        self.clip_coefficient = float(clip_coefficient)
        self.clip_value_loss = bool(clip_value_loss)
        self.entropy_loss_coefficient = float(entropy_loss_coefficient)
        self.value_loss_coefficient = float(value_loss_coefficient)
        self.state_magnitude_coefficient = float(state_magnitude_coefficient)
        self.max_grad_norm = float(max_grad_norm)
        self.normalize_advantages = bool(normalize_advantages)

        # Keep optimizer and clipper separate to allow defining the optimizer
        # somewhere else in the future
        self.clipper = optax.clip_by_global_norm(self.max_grad_norm)
        clipper_state = self.clipper.init(eqx.filter(policy, eqx.is_inexact_array))
        self.clipper_index = eqx.nn.StateIndex(clipper_state)

        self.optimizer = optax.inject_hyperparams(optax.adam)(3e-4)
        opt_state = self.optimizer.init(eqx.filter(policy, eqx.is_inexact_array))
        self.opt_state_index = eqx.nn.StateIndex(opt_state)

    @staticmethod
    def ppo_loss(
        policy: AbstractActorCriticPolicy[Float, ActType, ObsType],
        rollout_buffer: RolloutBuffer[ActType, ObsType],
        normalize_advantages: bool,
        clip_coefficient: float,
        clip_value_loss: bool,
        value_loss_coefficient: float,
        state_magnitude_coefficient: float,
        entropy_loss_coefficient: float,
    ) -> tuple[Float[Array, ""], PPOStats]:
        _, values, log_probs, entropy = jax.vmap(policy.evaluate_action)(
            rollout_buffer.states, rollout_buffer.observations, rollout_buffer.actions
        )

        log_ratios = log_probs - rollout_buffer.log_probs
        ratios = jnp.exp(log_ratios)
        approx_kl = jnp.mean(ratios - log_ratios) - 1

        advantages = rollout_buffer.advantages
        if normalize_advantages:
            advantages = (advantages - jnp.mean(advantages)) / (
                jnp.std(advantages) + jnp.finfo(advantages.dtype).eps
            )

        policy_loss = -jnp.mean(
            jnp.minimum(
                advantages * ratios,
                advantages
                * jnp.clip(ratios, 1 - clip_coefficient, 1 + clip_coefficient),
            )
        )

        if clip_value_loss:
            clipped_values = rollout_buffer.values + jnp.clip(
                values - rollout_buffer.values,
                -clip_coefficient,
                clip_coefficient,
            )
            value_loss = (
                jnp.mean(
                    jnp.minimum(
                        jnp.square(values - rollout_buffer.returns),
                        jnp.square(clipped_values - rollout_buffer.returns),
                    )
                )
                / 2
            )
        else:
            value_loss = jnp.mean(jnp.square(values - rollout_buffer.returns)) / 2

        entropy_loss = -jnp.mean(entropy)

        # TODO: Add state magnitude loss
        # State loss is proportional to the squared norm of the latent state of the policy
        state_magnitude_loss = jnp.array(0.0)

        loss = (
            policy_loss
            + value_loss * value_loss_coefficient
            + state_magnitude_loss * state_magnitude_coefficient
            + entropy_loss * entropy_loss_coefficient
        )

        return loss, PPOStats(
            approx_kl,
            loss,
            policy_loss,
            value_loss,
            entropy_loss,
            state_magnitude_loss,
        )

    ppo_loss_grad = staticmethod(eqx.filter_value_and_grad(ppo_loss, has_aux=True))

    def train_batch(
        self,
        state: eqx.nn.State,
        policy: AbstractActorCriticPolicy[Float, ActType, ObsType],
        rollout_buffer: RolloutBuffer[ActType, ObsType],
    ) -> tuple[
        eqx.nn.State, AbstractActorCriticPolicy[Float, ActType, ObsType], PPOStats
    ]:
        """
        Train the policy for one batch using the rollout buffer.

        Assumes that the rollout buffer is a single batch of data.
        """
        (_, stats), grads = self.ppo_loss_grad(
            policy,
            rollout_buffer,
            self.normalize_advantages,
            self.clip_coefficient,
            self.clip_value_loss,
            self.value_loss_coefficient,
            self.state_magnitude_coefficient,
            self.entropy_loss_coefficient,
        )

        clipper_state = state.get(self.clipper_index)
        updates, clipper_state = self.clipper.update(grads, clipper_state)
        state = state.set(self.clipper_index, clipper_state)

        opt_state = state.get(self.opt_state_index)

        updates, new_opt_state = self.optimizer.update(updates, opt_state)
        state = state.set(self.opt_state_index, new_opt_state)

        policy = eqx.apply_updates(policy, updates)

        return state, policy, stats

    def train_epoch(
        self,
        state: eqx.nn.State,
        policy: AbstractActorCriticPolicy[Float, ActType, ObsType],
        rollout_buffer: RolloutBuffer[ActType, ObsType],
        *,
        key: Key,
    ):
        """
        Train the policy for one epoch using the rollout buffer.

        One epoch consists of multiple mini-batches.
        """

        def batch_scan(
            carry: tuple[
                eqx.nn.State, AbstractActorCriticPolicy[Float, ActType, ObsType]
            ],
            rollout_buffer: RolloutBuffer[ActType, ObsType],
        ) -> tuple[
            tuple[eqx.nn.State, AbstractActorCriticPolicy[Float, ActType, ObsType]],
            PPOStats,
        ]:
            state, policy = carry
            state, policy, stats = self.train_batch(state, policy, rollout_buffer)
            return (state, policy), stats

        (state, policy), stats = filter_scan(
            batch_scan,
            (state, policy),
            rollout_buffer.batches(self.batch_size, key=key),
        )

        stats = jax.tree.map(jnp.mean, stats)

        return state, policy, stats

    @staticmethod
    def explained_variance(
        returns: Float[Array, ""], values: Float[Array, ""]
    ) -> Float[Array, ""]:
        variance = jnp.var(returns)
        explained_variance = 1 - jnp.var(returns - values) / (
            variance + jnp.finfo(returns.dtype).eps
        )
        return explained_variance

    def train(
        self,
        state: eqx.nn.State,
        policy: AbstractActorCriticPolicy[Float, ActType, ObsType],
        rollout_buffer: RolloutBuffer[ActType, ObsType],
        *,
        key: Key,
    ) -> tuple[
        eqx.nn.State,
        AbstractActorCriticPolicy[Float, ActType, ObsType],
        dict[str, Scalar],
    ]:
        def epoch_scan(
            carry: tuple[
                eqx.nn.State, AbstractActorCriticPolicy[Float, ActType, ObsType], Key
            ],
            _,
        ) -> tuple[
            tuple[
                eqx.nn.State, AbstractActorCriticPolicy[Float, ActType, ObsType], Key
            ],
            PPOStats,
        ]:
            state, policy, key = carry
            epoch_key, carry_key = jr.split(key, 2)
            state, policy, stats = self.train_epoch(
                state, policy, rollout_buffer, key=epoch_key
            )
            return (state, policy, carry_key), stats

        (state, policy, _), stats = filter_scan(
            epoch_scan, (state, policy, key), length=self.num_epochs
        )

        stats = jax.tree.map(jnp.mean, stats)
        explained_variance = self.explained_variance(
            rollout_buffer.returns, rollout_buffer.values
        )
        log = {
            "approx_kl": stats.approx_kl,
            "loss": stats.total_loss,
            "policy_loss": stats.policy_loss,
            "value_loss": stats.value_loss,
            "entropy_loss": stats.entropy_loss,
            "state_magnitude_loss": stats.state_magnitude_loss,
            "explained_variance": explained_variance,
        }

        return state, policy, log

    @classmethod
    def load(cls, path: str) -> PPO[ActType, ObsType]:
        raise NotImplementedError

    def save(self, path: str) -> None:
        raise NotImplementedError



from __future__ import annotations

from typing import cast

import equinox as eqx
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Float, Integer, Key, Real

from lerax.distribution import (
    AbstractDistribution,
    Categorical,
    SquashedMultivariateNormalDiag,
    SquashedNormal,
)
from lerax.env import AbstractEnvLike
from lerax.model import MLP
from lerax.space import Box, Discrete

from .actor_critic import AbstractActorCriticPolicy


class MLPActorCriticPolicy[
    FeatureType: Array,
    ActType: (Float[Array, " dims"], Integer[Array, ""]),
    ObsType: Real[Array, "..."],
](AbstractActorCriticPolicy[FeatureType, ActType, ObsType]):
    """
    Actor–critic policy with MLP components.
    """

    action_space: Box | Discrete
    observation_space: Box | Discrete

    feature_extractor: MLP
    value_model: MLP
    action_model: MLP
    log_std: Float[Array, " action_size"]

    def __init__(
        self,
        env: AbstractEnvLike[ActType, ObsType],
        *,
        feature_size: int = 64,
        feature_width: int = 64,
        feature_depth: int = 0,
        value_width: int = 64,
        value_depth: int = 2,
        action_width: int = 64,
        action_depth: int = 2,
        log_std_init: float = 0.0,
        key: Key,
    ):
        if isinstance(env.action_space, Discrete):
            act_size = int(env.action_space.n)
            self.log_std = jnp.array([], dtype=float)
        elif isinstance(env.action_space, Box):
            if env.action_space.shape:
                act_size = int(jnp.prod(jnp.asarray(env.action_space.shape)))
                self.log_std = jnp.full((act_size,), log_std_init, dtype=float)
            else:
                act_size = "scalar"
                self.log_std = jnp.array(log_std_init, dtype=float)
        else:
            raise NotImplementedError(
                f"Action space {type(env.action_space)} not supported."
            )

        if not isinstance(env.observation_space, (Discrete, Box)):
            raise NotImplementedError(
                f"Observation space {type(env.observation_space)} not supported."
            )

        self.action_space = env.action_space
        self.observation_space = env.observation_space

        feat_key, val_key, act_key = jr.split(key, 3)

        self.feature_extractor = MLP(
            in_size=int(jnp.array(self.observation_space.flat_size)),
            out_size=feature_size,
            width_size=feature_width,
            depth=feature_depth,
            key=feat_key,
        )

        self.value_model = MLP(
            in_size=feature_size,
            out_size="scalar",
            width_size=value_width,
            depth=value_depth,
            key=val_key,
        )

        self.action_model = MLP(
            in_size=feature_size,
            out_size=act_size,
            width_size=action_width,
            depth=action_depth,
            key=act_key,
        )

    def extract_features(
        self, state: eqx.nn.State, observation: ObsType
    ) -> tuple[eqx.nn.State, FeatureType]:
        """Extract features from an observation."""
        features = self.feature_extractor(jnp.ravel(observation))
        return state, cast(FeatureType, features)

    def action_dist_from_features(
        self, state: eqx.nn.State, features: FeatureType
    ) -> tuple[
        eqx.nn.State,
        AbstractDistribution[ActType],
    ]:
        """Return an action distribution from features."""
        action_mean = self.action_model(features)

        if isinstance(self.action_space, Discrete):
            action_dist: AbstractDistribution[ActType] = cast(
                AbstractDistribution[ActType],
                Categorical(logits=action_mean),
            )
        elif isinstance(self.action_space, Box):
            if self.action_space.shape == ():
                base_dist = SquashedNormal(
                    loc=action_mean,
                    scale=jnp.exp(self.log_std),
                )
            else:
                base_dist = SquashedMultivariateNormalDiag(
                    loc=action_mean,
                    scale_diag=jnp.exp(self.log_std),
                )
            action_dist = cast(AbstractDistribution[ActType], base_dist)
        else:
            raise NotImplementedError(
                f"Action space {type(self.action_space)} not supported."
            )

        return state, action_dist

    def value_from_features(
        self, state: eqx.nn.State, features: FeatureType
    ) -> tuple[eqx.nn.State, Float[Array, ""]]:
        """Return a value from features."""
        value = self.value_model(features)
        return state, value

    def reset(self, state: eqx.nn.State) -> eqx.nn.State:
        """Reset the policy state."""
        return state



from __future__ import annotations

from typing import ClassVar, Literal

import equinox as eqx
import gymnasium as gym
import jax
import numpy as np
from jax import numpy as jnp
from jax import random as jr
from jax.debug import callback as debug_callback
from jax.experimental import io_callback
from jaxtyping import Array, Bool, Float, Key

from lerax.env import AbstractEnv
from lerax.space import AbstractSpace, Box, Discrete


def gym_space_to_lerax_space(space: gym.Space) -> Box | Discrete:
    if isinstance(space, gym.spaces.Discrete):
        return Discrete(n=space.n)
    elif isinstance(space, gym.spaces.Box):
        return Box(low=space.low, high=space.high, shape=space.shape)
    else:
        raise NotImplementedError(f"Space type {type(space)} not supported")


def lerax_to_gym_space(space: AbstractSpace) -> gym.Space:
    if isinstance(space, Discrete):
        return gym.spaces.Discrete(int(space.n), start=int(space.start))
    elif isinstance(space, Box):
        return gym.spaces.Box(
            low=np.asarray(space.low),
            high=np.asarray(space.high),
        )
    else:
        raise NotImplementedError(f"Space type {type(space)} not supported")


def jax_to_numpy(x):
    if isinstance(x, jnp.ndarray):
        return np.asarray(x)
    return x


def to_numpy_tree(x):
    return jax.tree.map(jax_to_numpy, x)


class GymnasiumEnv(AbstractEnv[Array, Array]):
    """
    Wrapper of a Gymnasium environment to make it compatible with Lerax.

    Uses jax's io_callback to wrap the env's reset and step functions.
    In general, this will be slower than a native JAX environment,
    Also removes the info dict returned by Gymnasium envs.
    """

    name: ClassVar[str] = "GymnasiumEnv"

    state_index: eqx.nn.StateIndex[None]
    action_space: Box | Discrete
    observation_space: Box | Discrete

    env: gym.Env = eqx.field(static=True)

    def __init__(self, env: gym.Env):
        self.env = env
        self.action_space = gym_space_to_lerax_space(env.action_space)
        self.observation_space = gym_space_to_lerax_space(env.observation_space)
        self.state_index = eqx.nn.StateIndex(None)

    def _reset(self, *args, **kwargs):
        # TODO: Log a warning that the info dict is discarded
        kwargs["seed"] = int(kwargs["seed"])
        obs, _ = self.env.reset(*args, **kwargs)
        obs = jnp.asarray(obs, dtype=self.observation_space.dtype)
        return obs, {}

    def reset(
        self, state: eqx.nn.State, *args, key: Key, **kwargs
    ) -> tuple[eqx.nn.State, Array, dict]:
        # TODO: Determine if we want to pass a seed or not
        # I think it's a nice perk to increase reproducibility but it might
        # be unexpected for some users
        kwargs["seed"] = kwargs.get(
            "seed",
            jr.randint(key, (), 0, jnp.iinfo(jnp.int32).max),
        )

        return state, *io_callback(
            self._reset,
            (self.observation_space.canonical(), {}),
            *args,
            ordered=True,
            **kwargs,
        )

    def _step(self, action: Array):
        obs, reward, terminated, truncated, _ = self.env.step(np.asarray(action))

        return (
            jnp.asarray(obs, dtype=self.observation_space.dtype),
            jnp.asarray(reward),
            jnp.asarray(terminated),
            jnp.asarray(truncated),
            {},
        )

    def step(
        self, state: eqx.nn.State, action: Array, *, key: Key
    ) -> tuple[
        eqx.nn.State, Array, Float[Array, ""], Bool[Array, ""], Bool[Array, ""], dict
    ]:
        return state, *io_callback(
            self._step,
            (
                self.observation_space.canonical(),
                jnp.array(0.0),
                jnp.array(False),
                jnp.array(False),
                {},
            ),
            action,
            ordered=True,
        )

    def render(self, state: eqx.nn.State):
        # TODO: Log a warning that render is bypassed
        pass

    def close(self):
        debug_callback(self.env.close, ordered=True)


class LeraxEnv(gym.Env):
    """
    Wrapper of an Lerax environment to make it compatible with Gymnasium.

    Executes the Lerax env directly (Python side). Keeps an internal eqx state and PRNG.
    """

    metadata: dict = {"render_modes": ["human"]}

    action_space: gym.Space
    observation_space: gym.Space

    render_mode: str | None = None

    env: AbstractEnv
    state: eqx.nn.State
    key: Key

    def __init__(
        self,
        env: AbstractEnv,
        state: eqx.nn.State,
        render_mode: Literal["human"] | None = None,
    ):
        self.env = env
        self.state = state

        self.key = jr.key(0)

        self.action_space = lerax_to_gym_space(env.action_space)
        self.observation_space = lerax_to_gym_space(env.observation_space)

        self.render_mode = render_mode
        # TODO: Actually handle rendering

    def reset(self, *, seed: int | None = None, options: dict | None = None):
        if seed is not None:
            self.key = jr.key(int(seed))

        self.key, reset_key = jr.split(self.key)
        self.state, obs, info = self.env.reset(self.state, key=reset_key)
        return jax_to_numpy(obs), to_numpy_tree(info)

    def step(self, action):
        if isinstance(self.env.action_space, Discrete):
            act = jnp.asarray(action, dtype=self.env.action_space.dtype)
        elif isinstance(self.env.action_space, Box):
            act = jnp.asarray(action, dtype=self.env.action_space.dtype)
        else:
            raise NotImplementedError(
                f"Unsupported action space {type(self.env.action_space)}."
            )

        self.key, step_key = jr.split(self.key)
        self.state, obs, rew, term, trunc, info = self.env.step(
            self.state, act, key=step_key
        )

        return (
            jax_to_numpy(obs),
            float(jnp.asarray(rew)),
            bool(jnp.asarray(term)),
            bool(jnp.asarray(trunc)),
            to_numpy_tree(info),
        )

    def render(self):
        self.env.render(self.state)

    def close(self):
        self.env.close()



from __future__ import annotations

from abc import abstractmethod
from functools import partial
from typing import Callable

import equinox as eqx
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Bool, Float, Key

from lerax.env import AbstractEnvLike
from lerax.space import AbstractSpace, Box

from .base_wrapper import (
    AbstractNoActionSpaceWrapper,
    AbstractNoRenderOrCloseWrapper,
)
from .utils import rescale_box


class AbstractTransformObservationWrapper[WrapperObsType, ActType, ObsType](
    AbstractNoRenderOrCloseWrapper[ActType, WrapperObsType, ActType, ObsType],
    AbstractNoActionSpaceWrapper[WrapperObsType, ActType, ObsType],
):
    """Base class for environment observation wrappers"""

    env: eqx.AbstractVar[AbstractEnvLike[ActType, ObsType]]

    def reset(
        self, state: eqx.nn.State, *, key: Key
    ) -> tuple[eqx.nn.State, WrapperObsType, dict]:
        env_key, wrapper_key = jr.split(key, 2)

        substate = state.substate(self.env)
        substate, obs, info = self.env.reset(substate, key=env_key)
        state, obs = self.observation(state, obs, key=wrapper_key)

        state = state.update(substate)

        return state, obs, info

    def step(self, state: eqx.nn.State, action: ActType, *, key: Key) -> tuple[
        eqx.nn.State,
        WrapperObsType,
        Float[Array, ""],
        Bool[Array, ""],
        Bool[Array, ""],
        dict,
    ]:
        env_key, wrapper_key = jr.split(key, 2)

        env_state = state.substate(self.env)
        env_state, obs, reward, termination, truncation, info = self.env.step(
            env_state, action, key=env_key
        )
        state, obs = self.observation(state, obs, key=wrapper_key)

        state = state.update(env_state)

        return state, obs, reward, termination, truncation, info

    @abstractmethod
    def observation(
        self, state: eqx.nn.State, obs: ObsType, *, key: Key
    ) -> tuple[eqx.nn.State, WrapperObsType]:
        """Transform the wrapped environment observation"""


class AbstractPureObservationWrapper[WrapperObsType, ActType, ObsType](
    AbstractTransformObservationWrapper[WrapperObsType, ActType, ObsType]
):
    """
    Apply a pure function to every observation that leaves the environment.
    """

    env: eqx.AbstractVar[AbstractEnvLike[ActType, ObsType]]
    func: eqx.AbstractVar[Callable[[ObsType], WrapperObsType]]
    observation_space: eqx.AbstractVar[AbstractSpace[WrapperObsType]]

    def observation(
        self, state: eqx.nn.State, obs: ObsType, *, key: Key
    ) -> tuple[eqx.nn.State, WrapperObsType]:
        return state, self.func(obs)


class ClipObservation[ActType](
    AbstractPureObservationWrapper[Float[Array, " ..."], ActType, Float[Array, " ..."]],
):
    """
    Clips every observation to the environment's observation space.
    """

    env: AbstractEnvLike[ActType, Float[Array, " ..."]]
    func: Callable[[Float[Array, " ..."]], Float[Array, " ..."]]
    observation_space: Box

    def __init__(self, env: AbstractEnvLike[ActType, Float[Array, " ..."]]):
        if not isinstance(env.observation_space, Box):
            raise ValueError(
                "ClipObservation only supports `Box` observation spaces "
                f" not {type(env.observation_space)}"
            )

        self.env = env
        self.func = partial(
            jnp.clip,
            min=env.observation_space.low,
            max=env.observation_space.high,
        )
        self.observation_space = env.observation_space


class RescaleObservation[ActType](
    AbstractPureObservationWrapper[Float[Array, " ..."], ActType, Float[Array, " ..."]],
):
    """Affinely rescale a box observation to a different range"""

    env: AbstractEnvLike[ActType, Float[Array, " ..."]]
    func: Callable[[Float[Array, " ..."]], Float[Array, " ..."]]
    observation_space: Box

    def __init__(
        self,
        env: AbstractEnvLike[ActType, Float[Array, " ..."]],
        min: Float[Array, " ..."] = jnp.array(-1.0),
        max: Float[Array, " ..."] = jnp.array(1.0),
    ):
        if not isinstance(env.observation_space, Box):
            raise ValueError(
                "RescaleObservation only supports `Box` observation spaces "
                f" not {type(env.action_space)}"
            )

        new_box, forward, _ = rescale_box(env.observation_space, min, max)

        self.env = env
        self.func = forward
        self.observation_space = new_box


class FlattenObservation[ActType, ObsType](
    AbstractPureObservationWrapper[Float[Array, " flat"], ActType, ObsType]
):
    """Flatten the observation space into a 1-D array."""

    env: AbstractEnvLike[ActType, ObsType]
    func: Callable[[ObsType], Float[Array, " flat"]]
    observation_space: Box

    def __init__(self, env: AbstractEnvLike[ActType, ObsType]):
        self.env = env
        self.func = self.env.observation_space.flatten_sample
        self.observation_space = Box(
            -jnp.inf,
            jnp.inf,
            shape=(int(jnp.asarray(self.env.observation_space.flat_size)),),
        )



from __future__ import annotations

# Disable pygame greeting message
import contextlib
import warnings

with contextlib.redirect_stdout(None):
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        import pygame
        from pygame import gfxdraw

from jax import numpy as jnp
from jaxtyping import ArrayLike, Float

from .base_renderer import WHITE, AbstractRenderer, Color, Transform


class PygameRenderer(AbstractRenderer):
    """
    PyGame renderer implementation.
    """

    transform: Transform

    width: int
    height: int

    screen: pygame.Surface
    background_color: Color

    def __init__(
        self,
        width: int,
        height: int,
        background_color: Color = WHITE,
        transform: Transform | None = None,
    ) -> None:
        self.width = width
        self.height = height
        self.background_color = background_color

        if transform is None:
            self.transform = Transform(
                width=width,
                height=height,
                scale=jnp.array(1.0),
                offset=jnp.array([0.0, 0.0]),
            )
        else:
            self.transform = transform

        pygame.init()
        self.screen = pygame.display.set_mode((self.width, self.height))

    def is_open(self) -> bool:
        return True

    def open(self):
        pass

    def close(self):
        pygame.display.quit()
        pygame.quit()

    def render(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.close()
                return
        pygame.display.flip()

    def clear(self):
        self.screen.fill(self._pg_color(self.background_color))

    @staticmethod
    def _pg_color(c: Color) -> pygame.Color:
        r, g, b = c.to_rgb255()
        return pygame.Color(r, g, b)

    def _to_px(self, point: Float[ArrayLike, "2"]) -> tuple[int, int]:
        return self.transform.world_to_px(
            jnp.asarray(point)
        ).tolist()  # pyright: ignore

    def _scale_x(self, length: Float[ArrayLike, ""]) -> int:
        return int(self.transform.scale_length(length))  # pyright: ignore

    def draw_circle(
        self, center: Float[ArrayLike, "2"], radius: Float[ArrayLike, ""], color: Color
    ):
        gfxdraw.aacircle(
            self.screen,
            *self._to_px(center),
            self._scale_x(radius),
            self._pg_color(color),
        )
        gfxdraw.filled_circle(
            self.screen,
            *self._to_px(center),
            self._scale_x(radius),
            self._pg_color(color),
        )

    def draw_line(
        self,
        start: Float[ArrayLike, "2"],
        end: Float[ArrayLike, "2"],
        color: Color,
        width: Float[ArrayLike, ""] = 1,
    ):
        start = jnp.asarray(start)
        end = jnp.asarray(end)

        # Uses rectangle to work around lack of width in aaline
        norm = jnp.array([start[1] - end[1], end[0] - start[0]])  # pyright: ignore
        norm = norm / jnp.linalg.norm(norm) * (width / 2)

        p1 = start + norm
        p2 = start - norm
        p3 = end - norm
        p4 = end + norm

        self.draw_polygon(jnp.asarray([p1, p2, p3, p4]), color)

    def draw_rect(
        self,
        center: Float[ArrayLike, "2"],
        w: Float[ArrayLike, ""],
        h: Float[ArrayLike, ""],
        color: Color,
    ):
        top_left = self._to_px(
            jnp.asarray(center) - jnp.array([w / 2, -h / 2])  # pyright: ignore
        )
        width_hight = (self._scale_x(w), self._scale_x(h))

        rect = pygame.Rect(top_left, width_hight)

        gfxdraw.box(self.screen, rect, self._pg_color(color))

    def draw_polygon(self, points: Float[ArrayLike, "num 2"], color: Color):
        pts = [self._to_px(point) for point in points]  # pyright: ignore
        if len(pts) >= 3:
            gfxdraw.aapolygon(self.screen, pts, self._pg_color(color))
            gfxdraw.filled_polygon(self.screen, pts, self._pg_color(color))
        else:
            raise ValueError("Need at least 3 points to draw a polygon.")

    def draw_text(
        self,
        center: Float[ArrayLike, "2"],
        text: str,
        color: Color,
        size: Float[ArrayLike, ""] = 12,
    ):
        if not pygame.font.get_init():
            pygame.font.init()
        font = pygame.font.SysFont(None, int(size))  # pyright: ignore
        surf = font.render(text, True, self._pg_color(color))
        px, py = self._to_px(center)
        self.screen.blit(surf, (px, py))

    def draw_polyline(
        self,
        points: Float[ArrayLike, "num 2"],
        color: Color,
    ):
        if len(points) >= 2:  # pyright: ignore
            pygame.draw.aalines(
                self.screen,
                points=[self._to_px(p) for p in points],  # pyright: ignore
                color=self._pg_color(color),
                closed=False,
            )
        else:
            raise ValueError("Need at least 2 points to draw a polyline.")

    def draw_ellipse(
        self,
        center: Float[ArrayLike, "2"],
        w: Float[ArrayLike, ""],
        h: Float[ArrayLike, ""],
        color: Color,
    ):
        px, py = self._to_px(center)
        rx = max(1, int(self._scale_x(w) / 2))
        ry = max(1, int(self._scale_x(h) / 2))

        gfxdraw.aaellipse(self.screen, px, py, rx, ry, color=self._pg_color(color))

    def as_array(self) -> Float[ArrayLike, "height width 3"]:
        arr = pygame.surfarray.array3d(self.screen).transpose(1, 0, 2).copy()
        return arr



from __future__ import annotations

from typing import Any

from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key

from .base_space import AbstractSpace
from .utils import try_cast


class Box(AbstractSpace[Float[Array, " ..."]]):
    """
    A space of continuous values.

    A continuous closed set of floats.
    """

    _shape: tuple[int, ...]
    high: Float[Array, " ..."]
    low: Float[Array, " ..."]

    def __init__(
        self,
        low: Float[ArrayLike, " ..."],
        high: Float[ArrayLike, " ..."],
        shape: tuple[int, ...] | None = None,
    ):
        low = jnp.asarray(low, dtype=float)
        high = jnp.asarray(high, dtype=float)
        if shape is None:
            low, high = jnp.broadcast_arrays(low, high)
            shape = low.shape
            # TODO: Add warning if both shapes change

        assert low.shape == high.shape, "Box low and high must have the same shape"

        self._shape = shape
        self.low = jnp.broadcast_to(low, shape)
        self.high = jnp.broadcast_to(high, shape)

    @property
    def shape(self) -> tuple[int, ...]:
        return self._shape

    def canonical(self) -> Float[Array, " ..."]:
        return (self.low + self.high) / 2

    def sample(self, key: Key) -> Float[Array, " ..."]:
        bounded_key, unbounded_key, upper_bounded_key, lower_bounded_key = jr.split(
            key, 4
        )

        bounded_above = jnp.isfinite(self.high)
        bounded_below = jnp.isfinite(self.low)

        bounded = bounded_above & bounded_below
        unbounded = ~bounded_above & ~bounded_below
        upper_bounded = ~bounded_below & bounded_above
        lower_bounded = bounded_below & ~bounded_above

        sample = jnp.empty(self._shape, dtype=self.low.dtype)

        sample = jnp.where(
            bounded,
            jr.uniform(bounded_key, self._shape, minval=self.low, maxval=self.high),
            sample,
        )

        sample = jnp.where(unbounded, jr.normal(unbounded_key, self._shape), sample)

        sample = jnp.where(
            upper_bounded,
            self.high - jr.exponential(upper_bounded_key, self._shape),
            sample,
        )

        sample = jnp.where(
            lower_bounded,
            self.low + jr.exponential(lower_bounded_key, self._shape),
            sample,
        )

        return sample

    def contains(self, x: Any) -> Bool[ArrayLike, ""]:
        x = try_cast(x)
        if x is None:
            return False

        if x.shape != self._shape:
            return False

        return jnp.logical_and(jnp.all(x >= self.low), jnp.all(x <= self.high))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Box):
            return False

        return bool(jnp.array_equal(self.low, other.low)) and bool(
            jnp.array_equal(self.high, other.high)
        )

    def __repr__(self) -> str:
        return f"Box(low={self.low}, high={self.high})"

    def __hash__(self) -> int:
        return hash((self.low.tobytes(), self.high.tobytes()))

    def flatten_sample(self, sample: Float[Array, " ..."]) -> Float[Array, " size"]:
        return jnp.asarray(sample, dtype=float).ravel()

    @property
    def flat_size(self) -> Int[ArrayLike, ""]:
        return jnp.prod(jnp.asarray(self._shape)).astype(int)

    @property
    def dtype(self) -> jnp.dtype:
        return self.low.dtype



from __future__ import annotations

from distreqx import bijectors, distributions
from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Bool, Float, Integer

from .base_distribution import AbstractDistribution, AbstractTransformedDistribution


class Bernoulli(AbstractDistribution[Bool[Array, " dims"]]):

    distribution: distributions.Bernoulli

    def __init__(
        self,
        logits: Float[ArrayLike, " dims"] | None = None,
        probs: Float[ArrayLike, " dims"] | None = None,
    ):
        logits = jnp.asarray(logits) if logits is not None else None
        probs = jnp.asarray(probs) if probs is not None else None
        self.distribution = distributions.Bernoulli(logits=logits, probs=probs)

    @property
    def logits(self) -> Float[Array, " dims"]:
        return self.distribution.logits

    @property
    def probs(self) -> Float[Array, " dims"]:
        return self.distribution.probs


class Categorical(AbstractDistribution[Integer[Array, ""]]):

    distribution: distributions.Categorical

    def __init__(
        self,
        logits: Float[ArrayLike, " dims"] | None = None,
        probs: Float[ArrayLike, " dims"] | None = None,
    ):
        logits = jnp.asarray(logits) if logits is not None else None
        probs = jnp.asarray(probs) if probs is not None else None

        self.distribution = distributions.Categorical(logits=logits, probs=probs)

    @property
    def logits(self) -> Float[Array, " dims"]:
        return self.distribution.logits

    @property
    def probs(self) -> Float[Array, " dims"]:
        return self.distribution.probs


class Normal(AbstractDistribution[Float[Array, " dims"]]):

    distribution: distributions.Normal

    def __init__(
        self,
        loc: Float[ArrayLike, " dims"],
        scale: Float[ArrayLike, " dims"],
    ):
        loc = jnp.asarray(loc)
        scale = jnp.asarray(scale)

        if loc.shape != scale.shape:
            raise ValueError("loc and scale must have the same shape.")

        self.distribution = distributions.Normal(loc=loc, scale=scale)

    @property
    def loc(self) -> Float[Array, " dims"]:
        return self.distribution.loc

    @property
    def scale(self) -> Float[Array, " dims"]:
        return self.distribution.scale


class SquashedNormal(AbstractTransformedDistribution[Float[Array, " dims"]]):

    distribution: distributions.Transformed

    def __init__(
        self,
        loc: Float[ArrayLike, " dims"],
        scale: Float[ArrayLike, " dims"],
        high: Float[ArrayLike, " dims"] | None = None,
        low: Float[ArrayLike, " dims"] | None = None,
    ):
        loc = jnp.asarray(loc)
        scale = jnp.asarray(scale)
        high = jnp.asarray(high) if high is not None else None
        low = jnp.asarray(low) if low is not None else None

        if loc.shape != scale.shape:
            raise ValueError("loc and scale must have the same shape.")

        normal = distributions.Normal(loc=loc, scale=scale)

        if high is not None or low is not None:
            assert (
                high is not None and low is not None
            ), "Both high and low must be provided for bounded squashing."
            sigmoid = bijectors.Sigmoid()
            affine = bijectors.ScalarAffine(scale=high - low, shift=low)
            chain = bijectors.Chain((sigmoid, affine))
            self.distribution = distributions.Transformed(normal, chain)
        else:
            tanh = bijectors.Tanh()
            self.distribution = distributions.Transformed(normal, tanh)

    @property
    def loc(self) -> Float[Array, " dims"]:
        assert isinstance(self.distribution._distribution, distributions.Normal)
        return self.distribution._distribution.loc

    @property
    def scale(self) -> Float[Array, " dims"]:
        assert isinstance(self.distribution._distribution, distributions.Normal)
        return self.distribution._distribution.scale


class MultivariateNormalDiag(AbstractDistribution[Float[Array, " dims"]]):

    distribution: distributions.MultivariateNormalDiag

    def __init__(
        self,
        loc: Float[ArrayLike, " dims"] | None = None,
        scale_diag: Float[ArrayLike, " dims"] | None = None,
    ):
        loc = jnp.asarray(loc) if loc is not None else None
        scale_diag = jnp.asarray(scale_diag) if scale_diag is not None else None

        if (loc is not None and scale_diag is not None) and (
            loc.shape != scale_diag.shape
        ):
            raise ValueError("loc and scale_diag must have the same shape.")

        self.distribution = distributions.MultivariateNormalDiag(
            loc=loc, scale_diag=scale_diag
        )

    @property
    def loc(self) -> Float[Array, " dims"]:
        return self.distribution.loc

    @property
    def scale_diag(self) -> Float[Array, " dims"]:
        return self.distribution.scale_diag


class SquashedMultivariateNormalDiag(
    AbstractTransformedDistribution[Float[Array, " dims"]]
):
    """Multivariate Normal with squashing bijector for bounded outputs."""

    distribution: distributions.Transformed

    def __init__(
        self,
        loc: Float[ArrayLike, " dims"],
        scale_diag: Float[ArrayLike, " dims"],
        high: Float[ArrayLike, " dims"] | None = None,
        low: Float[ArrayLike, " dims"] | None = None,
    ):
        """
        Initialize a SquashedMultivariateNormalDiag distribution.

        Either both high and low must be provided for bounded squashing or neither.
        If neither are provided, the distribution will use a Tanh bijector for squashing
        between -1 and 1.
        """
        loc = jnp.asarray(loc)
        scale_diag = jnp.asarray(scale_diag)
        high = jnp.asarray(high) if high is not None else None
        low = jnp.asarray(low) if low is not None else None

        mvn = distributions.MultivariateNormalDiag(loc=loc, scale_diag=scale_diag)

        if high is not None or low is not None:
            assert (
                high is not None and low is not None
            ), "Both high and low must be provided for bounded squashing."

            sigmoid = bijectors.Sigmoid()
            scale = bijectors.DiagLinear(high - low)
            shift = bijectors.Shift(low)
            chain = bijectors.Chain((sigmoid, scale, shift))
            self.distribution = distributions.Transformed(mvn, chain)
        else:
            tanh = bijectors.Tanh()
            self.distribution = distributions.Transformed(mvn, tanh)

    @property
    def loc(self) -> Float[Array, " dims"]:
        assert isinstance(
            self.distribution._distribution, distributions.MultivariateNormalDiag
        )
        return self.distribution._distribution.loc

    @property
    def scale_diag(self) -> Float[Array, " dims"]:
        assert isinstance(
            self.distribution._distribution, distributions.MultivariateNormalDiag
        )
        return self.distribution._distribution.scale_diag



from __future__ import annotations

from typing import ClassVar

import equinox as eqx
from jax import lax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Bool, Float, Key

from lerax.space import Box

from .base_env import AbstractEnv


class ContinuousMountainCar(AbstractEnv[Float[Array, ""], Float[Array, "2"]]):
    name: ClassVar[str] = "ContinuousMountainCar"

    state_index: eqx.nn.StateIndex[Float[Array, "2"]]
    action_space: Box
    observation_space: Box

    min_action: float
    max_action: float
    min_position: float
    max_position: float
    max_speed: float
    goal_position: float
    goal_velocity: float

    power: float

    low: Float[Array, "2"]
    high: Float[Array, "2"]

    def __init__(self, goal_velocity: float = 0):
        self.min_action = -1.0
        self.max_action = 1.0
        self.min_position = -1.2
        self.max_position = 0.6
        self.max_speed = 0.07
        self.goal_position = 0.5
        self.goal_velocity = goal_velocity
        self.power = 0.0015

        self.low = jnp.array([self.min_position, -self.max_speed])
        self.high = jnp.array([self.max_position, self.max_speed])

        self.action_space = Box(self.min_action, self.max_action)
        self.observation_space = Box(self.low, self.high)

        self.state_index = eqx.nn.StateIndex(jnp.zeros(2))

    def reset(
        self, state: eqx.nn.State, *, key: Key, low: float = -0.6, high: float = -0.4
    ) -> tuple[eqx.nn.State, Float[Array, "2"], dict]:
        position = jr.uniform(key, minval=low, maxval=high)
        velocity = 0.0

        state_vals = jnp.asarray([position, velocity])
        state = state.set(self.state_index, state_vals)

        return state, state_vals, {}

    def step(self, state: eqx.nn.State, action: Float[Array, ""], *, key: Key) -> tuple[
        eqx.nn.State,
        Float[Array, "2"],
        Float[Array, ""],
        Bool[Array, ""],
        Bool[Array, ""],
        dict,
    ]:
        position, velocity = state.get(self.state_index)
        force = jnp.clip(action, self.min_action, self.max_action)

        velocity += self.power * force - jnp.cos(3 * position) * 0.0025
        velocity = jnp.clip(velocity, -self.max_speed, self.max_speed)

        position += velocity
        position = jnp.clip(position, self.min_position, self.max_position)

        velocity = lax.cond(
            jnp.logical_and(position == self.min_position, velocity < 0),
            lambda: jnp.array(0.0),
            lambda: velocity,
        )

        terminated = jnp.logical_and(
            position >= self.goal_position, velocity >= self.goal_velocity
        )

        reward = lax.cond(
            terminated,
            lambda: 100.0,
            lambda: 0.0,
        )
        reward -= force**2 * 0.1

        state_vals = jnp.asarray([position, velocity])
        state = state.set(self.state_index, state_vals)

        return state, state_vals, reward, terminated, jnp.array(False), {}

    def render(self, state: eqx.nn.State):
        raise NotImplementedError

    def close(self): ...



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx


class AbstractBuffer(eqx.Module):
    """Base class for buffers."""

    @property
    @abstractmethod
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the buffer."""



"""
Lerax models

Models take inputs and produce outputs, and may have state.
"""

from .base_model import (
    AbstractModel,
    AbstractStatefulModel,
)
from .flatten import Flatten
from .mlp import MLP
from .ncde import (
    AbstractNCDETerm,
    AbstractNeuralCDE,
    MLPNCDETerm,
    MLPNeuralCDE,
)
from .node import (
    AbstractNeuralODE,
    AbstractNODETerm,
    MLPNeuralODE,
    MLPNODETerm,
)

__all__ = [
    "AbstractModel",
    "AbstractStatefulModel",
    "Flatten",
    "AbstractNeuralODE",
    "AbstractNODETerm",
    "MLPNeuralODE",
    "MLPNODETerm",
    "AbstractNeuralCDE",
    "AbstractNCDETerm",
    "MLPNeuralCDE",
    "MLPNCDETerm",
    "MLP",
]



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
import optax
from jaxtyping import Key

from lerax.env import AbstractEnvLike
from lerax.policy import AbstractPolicy


class AbstractAlgorithm[ActType, ObsType](eqx.Module):
    """Base class for RL algorithms."""

    env: eqx.AbstractVar[AbstractEnvLike[ActType, ObsType]]
    policy: eqx.AbstractVar[AbstractPolicy]
    optimizer: eqx.AbstractVar[optax.GradientTransformation]
    opt_state_index: eqx.AbstractVar[eqx.nn.StateIndex[optax.OptState]]

    # TODO: Add support for callbacks
    @abstractmethod
    def learn(
        self,
        state: eqx.nn.State,
        total_timesteps: int,
        *,
        key: Key,
        show_progress_bar: bool = False,
        tb_log_name: str | None = None,
    ) -> tuple[eqx.nn.State, AbstractPolicy[ActType, ObsType]]:
        """Return a trained model."""

    @classmethod
    @abstractmethod
    def load(cls, path, *args, **kwargs) -> AbstractAlgorithm[ActType, ObsType]:
        """Load a model from a file."""

    @abstractmethod
    def save(self, path: str) -> None:
        """Save the model to a file."""



from __future__ import annotations

from typing import Any

from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key

from .base_space import AbstractSpace


class Tuple(AbstractSpace[tuple[Any, ...]]):
    """A cartesian product of spaces."""

    spaces: tuple[AbstractSpace, ...]

    def __init__(self, spaces: tuple[AbstractSpace, ...]):
        assert isinstance(spaces, tuple), "spaces must be a tuple"
        assert len(spaces) > 0, "spaces must be non-empty"
        assert all(
            isinstance(space, AbstractSpace) for space in spaces
        ), "spaces must be a tuple of AbstractSpace"

        self.spaces = spaces

    @property
    def shape(self) -> None:
        return None

    def canonical(self) -> tuple[Any, ...]:
        return tuple(space.canonical() for space in self.spaces)

    def sample(self, key: Key) -> tuple[Any, ...]:
        return tuple(
            space.sample(key)
            for space, key in zip(self.spaces, jr.split(key, len(self.spaces)))
        )

    def contains(self, x: Any) -> Bool[ArrayLike, ""]:
        if not isinstance(x, tuple):
            return False

        if len(x) != len(self.spaces):
            return False

        return jnp.array(
            space.contains(x_i) for space, x_i in zip(self.spaces, x)
        ).all()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Tuple):
            return False

        return all(
            space == other_space
            for space, other_space in zip(self.spaces, other.spaces)
        )

    def __repr__(self) -> str:
        return f"Tuple({', '.join(repr(space) for space in self.spaces)})"

    def __hash__(self) -> int:
        return hash(tuple(hash(space) for space in self.spaces))

    def flatten_sample(self, sample: tuple[Any, ...]) -> Float[Array, " size"]:
        parts = [
            subspace.flatten_sample(subsample)
            for subsample, subspace in zip(sample, self.spaces)
        ]
        return jnp.concatenate(parts)

    @property
    def flat_size(self) -> Int[ArrayLike, ""]:
        return jnp.array(space.flat_size for space in self.spaces).sum().astype(int)

    def __getitem__(self, index: int) -> AbstractSpace:
        return self.spaces[index]

    def __len__(self) -> int:
        return len(self.spaces)



from __future__ import annotations

from typing import Any

from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key

from .base_space import AbstractSpace
from .utils import try_cast


class Discrete(AbstractSpace[Int[Array, ""]]):
    """
    A space of finite discrete values.

    A finite closed set of integers.
    """

    _n: Int[Array, ""]
    start: Int[Array, ""]

    def __init__(self, n: Int[ArrayLike, ""], start: Int[ArrayLike, ""] = 0):
        assert n > 0, "n must be positive"  # pyright: ignore

        self._n = jnp.array(n, dtype=float)
        self.start = jnp.array(start, dtype=float)

    @property
    def n(self) -> Int[Array, ""]:
        return self._n

    @property
    def shape(self) -> tuple[int, ...]:
        return ()

    def canonical(self) -> Int[Array, ""]:
        return self.start

    def sample(self, key: Key) -> Int[Array, ""]:
        return jr.randint(key, shape=(), minval=self.start, maxval=self._n + self.start)

    def contains(self, x: Any) -> Bool[ArrayLike, ""]:
        x = try_cast(x)
        if x is None:
            return False

        if x.ndim != 0:
            return False
        x = x.squeeze()

        if jnp.logical_not(jnp.array_equal(x, jnp.floor(x))):
            return False

        return bool(self.start <= x < self._n + self.start)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Discrete):
            return False
        return bool(self._n == other._n) and bool(self.start == other.start)

    def __repr__(self) -> str:
        return f"Discrete({self._n}, start={self.start})"

    def __hash__(self) -> int:
        return hash((int(self._n), int(self.start)))

    def flatten_sample(self, sample: Int[Array, ""]) -> Float[Array, ""]:
        return jnp.asarray(sample, dtype=float)

    @property
    def flat_size(self) -> Int[ArrayLike, ""]:
        return jnp.array(1, dtype=int)

    @property
    def dtype(self) -> jnp.dtype:
        return self._n.dtype



from __future__ import annotations

from abc import abstractmethod
from typing import Callable, Literal

import equinox as eqx
from jax import nn as jnn
from jax import numpy as jnp
from jaxtyping import Array, Float, Key, ScalarLike


class AbstractNCDETerm(eqx.Module):
    """Base class for a neural ODE vector field term."""

    @abstractmethod
    def __call__(
        self, t: ScalarLike, z: Float[Array, " data_size"], args
    ) -> Float[Array, " data_size input_size"]:
        """Vector field value."""


class TensorMLP(eqx.Module):
    """
    Modification of the MLP class from Equinox to handle tensors as input and output.
    """

    in_shape: tuple[int, ...] | Literal["scalar"]
    out_shape: tuple[int, ...] | Literal["scalar"]
    mlp: eqx.nn.MLP

    def __init__(
        self,
        in_shape: tuple[int, ...] | Literal["scalar"],
        out_shape: tuple[int, ...] | Literal["scalar"],
        width_size: int,
        depth: int,
        activation: Callable[[Array], Array] = jnn.relu,
        final_activation: Callable[[Array], Array] = lambda x: x,
        use_bias: bool = True,
        use_final_bias: bool = True,
        dtype: jnp.dtype | None = None,
        *,
        key: Key,
    ) -> None:
        self.in_shape = in_shape
        self.out_shape = out_shape

        if in_shape == "scalar":
            in_size = "scalar"
        else:
            in_size = int(jnp.asarray(in_shape).prod())

        if out_shape == "scalar":
            out_size = "scalar"
        else:
            out_size = int(jnp.asarray(out_shape).prod())

        self.mlp = eqx.nn.MLP(
            in_size=in_size,
            out_size=out_size,
            width_size=width_size,
            depth=depth,
            activation=activation,
            final_activation=final_activation,
            use_bias=use_bias,
            use_final_bias=use_final_bias,
            dtype=dtype,
            key=key,
        )

    def __call__(self, x: Array) -> Array:
        if self.in_shape == "scalar":
            x = self.mlp(x)
        else:
            x = jnp.ravel(x)
            x = self.mlp(x)

        if self.out_shape != "scalar":
            x = jnp.reshape(x, self.out_shape)

        return x


class MLPNCDETerm(AbstractNCDETerm):
    add_time: bool
    tensor_mlp: TensorMLP

    def __init__(
        self,
        input_size: int,
        data_size: int,
        width_size: int,
        depth: int,
        activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = jnn.softplus,
        final_activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = jnn.tanh,
        *,
        key: Key,
        add_time: bool = True,
    ):
        self.add_time = add_time
        self.tensor_mlp = TensorMLP(
            in_shape=(data_size + int(add_time),),
            out_shape=(data_size, input_size),
            width_size=width_size,
            depth=depth,
            activation=activation,
            final_activation=final_activation,
            key=key,
        )

    def __call__(
        self, t: ScalarLike, z: Float[Array, " data_size"], args
    ) -> Float[Array, " data_size input_size"]:
        if self.add_time:
            return self.tensor_mlp(
                jnp.concatenate([z, jnp.expand_dims(t, axis=-1)], axis=-1)
            )
        else:
            return self.tensor_mlp(z)



from abc import abstractmethod

import equinox as eqx
from jax import numpy as jnp
from jaxtyping import ArrayLike, Float, Int


class Color(eqx.Module):
    """RGB in [0,1]. Helpers to convert to pygame-friendly formats."""

    r: Float[ArrayLike, ""]
    g: Float[ArrayLike, ""]
    b: Float[ArrayLike, ""]

    def to_tuple(
        self,
    ) -> tuple[Float[ArrayLike, ""], Float[ArrayLike, ""], Float[ArrayLike, ""]]:
        return (self.r, self.g, self.b)

    def to_rgb255(self) -> tuple[int, int, int]:
        r = int(float(jnp.clip(self.r, 0.0, 1.0)) * 255.0)
        g = int(float(jnp.clip(self.g, 0.0, 1.0)) * 255.0)
        b = int(float(jnp.clip(self.b, 0.0, 1.0)) * 255.0)
        return (r, g, b)

    def to_hex(self) -> str:
        r, g, b = self.to_rgb255()
        return f"#{r:02x}{g:02x}{b:02x}"

    @classmethod
    def from_hex(cls, hex_str: str) -> "Color":
        r = int(hex_str[1:3], 16) / 255.0
        g = int(hex_str[3:5], 16) / 255.0
        b = int(hex_str[5:7], 16) / 255.0
        return cls(r, g, b)

    @classmethod
    def from_rgb255(cls, r: int, g: int, b: int) -> "Color":
        return cls(r / 255.0, g / 255.0, b / 255.0)

    @classmethod
    def from_tuple(
        cls,
        rgb: tuple[Float[ArrayLike, ""], Float[ArrayLike, ""], Float[ArrayLike, ""]],
    ) -> "Color":
        return cls(*rgb)


WHITE = Color(1.0, 1.0, 1.0)
BLACK = Color(0.0, 0.0, 0.0)
GRAY = Color(0.78, 0.78, 0.78)
RED = Color(0.86, 0.24, 0.24)
GREEN = Color(0.24, 0.71, 0.29)
BLUE = Color(0.26, 0.53, 0.96)


class Transform(eqx.Module):
    """
    Affine mapping from world coords to screen pixels.

    Has a constant scale for all axis to avoid distortion.
    """

    width: Int[ArrayLike, ""]
    height: Int[ArrayLike, ""]

    scale: Float[ArrayLike, ""]
    offset: Float[ArrayLike, "2"]

    y_up: bool = True

    def world_to_px(self, point: Float[ArrayLike, "2"]) -> Int[ArrayLike, "2"]:
        """Return pixel coordinates corresponding to a point in world space."""
        pixel = jnp.asarray(point) * self.scale + self.offset

        if self.y_up:
            pixel = pixel.at[1].set(self.height - pixel[1])

        return pixel.astype(int)

    def scale_length(self, length: Float[ArrayLike, ""]) -> Float[ArrayLike, ""]:
        """Scale a length in world space to pixel space."""
        return length * self.scale


class AbstractRenderer(eqx.Module):
    """
    Renderer interface.

    All methods take coordinates in world space.
    Renderers are not necessarily thread-safe or JITtable.
    """

    transform: eqx.AbstractVar[Transform]

    @abstractmethod
    def is_open(self) -> bool: ...
    @abstractmethod
    def open(self): ...
    @abstractmethod
    def close(self): ...
    @abstractmethod
    def render(self): ...
    @abstractmethod
    def clear(self): ...

    @abstractmethod
    def draw_circle(
        self, center: Float[ArrayLike, "2"], radius: Float[ArrayLike, ""], color: Color
    ): ...
    @abstractmethod
    def draw_line(
        self,
        start: Float[ArrayLike, "2"],
        end: Float[ArrayLike, "2"],
        color: Color,
        width: Float[ArrayLike, ""] = 1,
    ): ...
    @abstractmethod
    def draw_rect(
        self,
        center: Float[ArrayLike, "2"],
        w: Float[ArrayLike, ""],
        h: Float[ArrayLike, ""],
        color: Color,
    ): ...
    @abstractmethod
    def draw_polygon(self, points: Float[ArrayLike, "num 2"], color: Color): ...
    @abstractmethod
    def draw_text(
        self,
        center: Float[ArrayLike, "2"],
        text: str,
        color: Color,
        size: Float[ArrayLike, ""] = 12,
    ): ...
    @abstractmethod
    def draw_polyline(
        self,
        points: Float[ArrayLike, "num 2"],
        color: Color,
    ): ...
    @abstractmethod
    def draw_ellipse(
        self,
        center: Float[ArrayLike, "2"],
        w: Float[ArrayLike, ""],
        h: Float[ArrayLike, ""],
        color: Color,
    ): ...
    @abstractmethod
    def as_array(self) -> Float[ArrayLike, "H W 3"]: ...



from __future__ import annotations

from typing import Any

from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key

from .base_space import AbstractSpace
from .utils import try_cast


class MultiDiscrete(AbstractSpace[Int[ArrayLike, " n"]]):
    """Cartesian product of discrete spaces."""

    ns: Int[Array, " n"]
    starts: Int[Array, " n"]

    def __init__(self, ns: tuple[int, ...], starts: tuple[int, ...] = (0,)):
        assert len(ns) > 0, "ns must be non-empty"
        starts = tuple(starts) if len(starts) > 0 else (0,) * len(ns)
        assert len(ns) == len(starts), "ns and starts must have the same length"
        assert all(n > 0 for n in ns), "all n must be positive"

        self.ns = jnp.array(ns, dtype=float)
        self.starts = jnp.array(starts, dtype=float)

    @property
    def shape(self) -> tuple[int, ...]:
        return (len(self.ns),)

    def canonical(self) -> Int[Array, " n"]:
        return self.starts

    def sample(self, key: Key) -> Int[Array, " n"]:
        return jr.randint(
            key, shape=self.shape, minval=self.starts, maxval=self.ns + self.starts
        )

    def contains(self, x: Any) -> Bool[ArrayLike, ""]:
        x = try_cast(x)
        if x is None:
            return False

        if x.shape != self.shape:
            return False

        if jnp.logical_not(jnp.array_equal(x, jnp.floor(x))):
            return False

        return jnp.all((self.starts <= x) & (x < self.ns + self.starts), axis=0)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MultiDiscrete):
            return False

        return bool(jnp.array_equal(self.ns, other.ns)) and bool(
            jnp.array_equal(self.starts, other.starts)
        )

    def __repr__(self) -> str:
        return f"MultiDiscrete({self.ns}, starts={self.starts})"

    def __hash__(self) -> int:
        return hash((self.ns.tobytes(), self.starts.tobytes()))

    def flatten_sample(self, sample: Int[ArrayLike, " n"]) -> Float[Array, " size"]:
        return jnp.asarray(sample, dtype=float).ravel()

    @property
    def flat_size(self) -> Int[ArrayLike, ""]:
        return jnp.array(len(self.ns), dtype=int)



from __future__ import annotations

import threading
from functools import partial, wraps
from typing import Any, Callable

import equinox as eqx
import jax
import numpy as np
from jax import lax
from jax import numpy as jnp


# TODO: Consider moving this to a standalone Equinox utilities library
def clone_state(state: eqx.nn.State) -> eqx.nn.State:
    """
    Clone an Equinox state.

    Equinox does not allow reuse of states. Cloning in this way bypasses this restriction.
    """
    leaves, treedef = jax.tree.flatten(state)
    state_clone = jax.tree.unflatten(treedef, leaves)
    return state_clone


# TODO: Consider moving this to a standalone Equinox utilities library
class _FilterScan(eqx.Module):

    @property
    def __wrapped__(self):
        return lax.scan

    def __call__(
        self,
        f,
        init,
        xs=None,
        length=None,
        reverse: bool = False,
        unroll: int | bool = 1,
        _split_transpose: bool = False,
    ):
        init_arr, static = eqx.partition(init, eqx.is_array)

        def _f(carry_arr, x):
            carry = eqx.combine(carry_arr, static)
            carry, y = f(carry, x)
            new_carry_arr, new_static = eqx.partition(carry, eqx.is_array)
            assert eqx.tree_equal(
                static, new_static
            ), "Non-array carry of filter_scan must not change."
            return new_carry_arr, y

        carry_arr, ys = lax.scan(
            f=_f,
            init=init_arr,
            xs=xs,
            length=length,
            reverse=reverse,
            unroll=unroll,
            _split_transpose=_split_transpose,
        )
        return eqx.combine(carry_arr, static), ys


filter_scan = eqx.module_update_wrapper(_FilterScan())


def debug_wrapper[**InType](
    func: Callable[InType, Any], ordered: bool = False, thread: bool | None = None
) -> Callable[InType, None]:
    """
    Return a JIT‑safe version of *func*.

    :param func: The function to wrap.
    :param ordered: If True, the callback will be executed in the order of the arguments
    :param thread: If True, the callback will be executed in a separate thread.
    """
    if thread is None:
        thread = not ordered

    if ordered and thread:
        # TODO: Add a warning or error here
        pass

    def _callback(*args: InType.args, **kwargs: InType.kwargs) -> None:
        if thread:
            threading.Thread(target=func, args=args, kwargs=kwargs, daemon=True).start()
        else:
            func(*args, **kwargs)

    @wraps(func)
    def wrapped(*args: InType.args, **kwargs: InType.kwargs) -> None:
        jax.debug.callback(_callback, *args, **kwargs, ordered=ordered)

    return wrapped


def debug_with_numpy_wrapper(
    func: Callable[..., Any], ordered: bool = False, thread: bool | None = None
) -> Callable[..., None]:
    """
    Like `debug_wrapper` but converts every jax.Array/`jnp.ndarray` argument
    to a plain `numpy.ndarray` before calling *func*.

    It is impossible with Python's current type system to express the transformation so
    parameter information is lost.
    """

    @partial(debug_wrapper, ordered=ordered, thread=thread)
    @wraps(func)
    def wrapped(*args, **kwargs) -> None:
        args, kwargs = jax.tree.map(
            lambda x: np.asarray(x) if isinstance(x, jnp.ndarray) else x, (args, kwargs)
        )
        func(*args, **kwargs)

    return wrapped


def debug_with_list_wrapper(
    func: Callable[..., Any], ordered: bool = False, thread: bool | None = None
) -> Callable[..., None]:
    """
    Like `debug_wrapper` but converts every jax.Array/`jnp.ndarray` argument
    to a plain list before calling *func*.

    It is impossible with Python's current type system to express the transformation so
    parameter information is lost.
    """

    @partial(debug_wrapper, ordered=ordered, thread=thread)
    @wraps(func)
    def wrapped(*args, **kwargs) -> None:
        args, kwargs = jax.tree.map(
            lambda x: (
                np.asarray(x).tolist()
                if isinstance(x, (jnp.ndarray, np.ndarray))
                else x
            ),
            (args, kwargs),
        )
        func(*args, **kwargs)

    return wrapped



from __future__ import annotations

from typing import Callable, Literal

import equinox as eqx
from jax import nn as jnn
from jaxtyping import Array, Float, Key

from .base_model import AbstractModel


class MLP(AbstractModel[[Float[Array, " in_size"]], Float[Array, " out_size"]]):
    """Wrapper around eqx.nn.MLP."""

    mlp: eqx.nn.MLP

    in_size: int | Literal["scalar"]
    out_size: int | Literal["scalar"]

    def __init__(
        self,
        in_size: int | Literal["scalar"],
        out_size: int | Literal["scalar"],
        width_size: int,
        depth: int,
        activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = jnn.relu,
        final_activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = lambda x: x,
        *,
        key: Key,
    ) -> None:
        self.in_size = in_size
        self.out_size = out_size

        self.mlp = eqx.nn.MLP(
            in_size=in_size,
            out_size=out_size,
            width_size=width_size,
            depth=depth,
            activation=activation,
            final_activation=final_activation,
            key=key,
        )

    def __call__(self, x: Float[Array, " in_size"]) -> Float[Array, " out_size"]:
        return self.mlp(x)



from __future__ import annotations

import equinox as eqx

from lerax.env import AbstractEnv, AbstractEnvLike
from lerax.space import AbstractSpace


class AbstractWrapper[WrapperActType, WrapperObsType, ActType, ObsType](
    AbstractEnvLike[WrapperActType, WrapperObsType]
):
    """Base class for environment wrappers"""

    env: eqx.AbstractVar[AbstractEnvLike[ActType, ObsType]]

    @property
    def unwrapped(self) -> AbstractEnv[ActType, ObsType]:
        """Return the unwrapped environment"""
        return self.env.unwrapped

    @property
    def name(self) -> str:
        """Return the name of the environment"""
        return self.env.name


class AbstractNoRenderWrapper[WrapperActType, WrapperObsType, ActType, ObsType](
    AbstractWrapper[WrapperActType, WrapperObsType, ActType, ObsType]
):
    """A wrapper that does not affect rendering"""

    env: eqx.AbstractVar[AbstractEnvLike[ActType, ObsType]]

    def render(self, state: eqx.nn.State):
        return self.env.render(state)


class AbstractNoCloseWrapper[WrapperActType, WrapperObsType, ActType, ObsType](
    AbstractWrapper[WrapperActType, WrapperObsType, ActType, ObsType]
):
    """A wrapper that does not affect closing"""

    env: eqx.AbstractVar[AbstractEnvLike[ActType, ObsType]]

    def close(self):
        return self.env.close()


class AbstractNoRenderOrCloseWrapper[WrapperActType, WrapperObsType, ActType, ObsType](
    AbstractNoRenderWrapper[WrapperActType, WrapperObsType, ActType, ObsType],
    AbstractNoCloseWrapper[WrapperActType, WrapperObsType, ActType, ObsType],
):
    """A wrapper that does not affect rendering or closing the environment"""

    env: eqx.AbstractVar[AbstractEnvLike[ActType, ObsType]]


class AbstractNoActionSpaceWrapper[WrapperObsType, ActType, ObsType](
    AbstractWrapper[ActType, WrapperObsType, ActType, ObsType]
):
    """A wrapper that does not affect the action space"""

    env: eqx.AbstractVar[AbstractEnvLike[ActType, ObsType]]

    @property
    def action_space(self) -> AbstractSpace[ActType]:
        return self.env.action_space


class AbstractNoObservationSpaceWrapper[WrapperActType, ActType, ObsType](
    AbstractWrapper[WrapperActType, ObsType, ActType, ObsType]
):
    """A wrapper that does not affect the observation space"""

    env: eqx.AbstractVar[AbstractEnvLike[ActType, ObsType]]

    @property
    def observation_space(self) -> AbstractSpace[ObsType]:
        return self.env.observation_space


class AbstractNoActionOrObservationSpaceWrapper[ActType, ObsType](
    AbstractNoActionSpaceWrapper[ObsType, ActType, ObsType],
    AbstractNoObservationSpaceWrapper[ActType, ActType, ObsType],
):
    """A wrapper that does not affect the action or observation space"""



from __future__ import annotations

from abc import abstractmethod
from typing import Callable

import equinox as eqx
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Key

from lerax.env import AbstractEnvLike

from .base_wrapper import (
    AbstractNoActionOrObservationSpaceWrapper,
    AbstractNoRenderOrCloseWrapper,
)


class AbstractTransformRewardWrapper[ActType, ObsType](
    AbstractNoRenderOrCloseWrapper[ActType, ObsType, ActType, ObsType],
    AbstractNoActionOrObservationSpaceWrapper[ActType, ObsType],
):
    """Base class for environment reward wrappers"""

    env: eqx.AbstractVar[AbstractEnvLike[ActType, ObsType]]

    def reset(
        self, state: eqx.nn.State, *, key: Key
    ) -> tuple[eqx.nn.State, ObsType, dict]:
        substate = state.substate(self.env)
        substate, obs, info = self.env.reset(substate, key=key)
        state = state.update(substate)

        return state, obs, info

    def step(
        self, state: eqx.nn.State, action: ActType, *, key: Key
    ) -> tuple[
        eqx.nn.State, ObsType, Float[Array, ""], Bool[Array, ""], Bool[Array, ""], dict
    ]:
        env_key, wrapper_key = jr.split(key, 2)

        env_state = state.substate(self.env)
        env_state, obs, reward, termination, truncation, info = self.env.step(
            env_state, action, key=env_key
        )
        state = state.update(env_state)

        reward = self.reward(state, reward, key=wrapper_key)

        return state, obs, reward, termination, truncation, info

    @abstractmethod
    def reward(
        self, state: eqx.nn.State, reward: Float[Array, ""], *, key: Key
    ) -> Float[Array, ""]:
        """Transform the reward from the wrapped environment"""


class AbstractPureTransformRewardWrapper[ActType, ObsType](
    AbstractTransformRewardWrapper[ActType, ObsType]
):
    """
    Apply a *pure* (stateless) function to every reward emitted by the wrapped
    environment.
    """

    env: eqx.AbstractVar[AbstractEnvLike[ActType, ObsType]]
    func: eqx.AbstractVar[Callable[[Float[Array, ""]], Float[Array, ""]]]

    def reward(
        self,
        state: eqx.nn.State,
        reward: Float[Array, ""],
        *,
        key: Key,
    ) -> Float[Array, ""]:
        return self.func(reward)


class ClipReward[ActType, ObsType](
    AbstractPureTransformRewardWrapper[ActType, ObsType]
):
    """
    Element-wise clip of rewards:  `reward ↦ clamp(min, max)`.
    """

    env: AbstractEnvLike[ActType, ObsType]
    func: Callable[[Float[Array, ""]], Float[Array, ""]]
    min: Float[Array, ""]
    max: Float[Array, ""]

    def __init__(
        self,
        env: AbstractEnvLike[ActType, ObsType],
        min: Float[ArrayLike, ""] = jnp.asarray(-1.0),
        max: Float[ArrayLike, ""] = jnp.asarray(1.0),
    ):
        self.env = env
        self.min = jnp.asarray(min)
        self.max = jnp.asarray(max)
        self.func = lambda r: jnp.clip(r, self.min, self.max)



from __future__ import annotations

from abc import abstractmethod
from typing import Callable

import equinox as eqx
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Bool, Float, Key

from lerax.env import AbstractEnvLike
from lerax.space import AbstractSpace, Box

from .base_wrapper import (
    AbstractNoObservationSpaceWrapper,
    AbstractNoRenderOrCloseWrapper,
)
from .utils import rescale_box


class AbstractTransformActionWrapper[WrapperActType, ActType, ObsType](
    AbstractNoRenderOrCloseWrapper[WrapperActType, ObsType, ActType, ObsType],
    AbstractNoObservationSpaceWrapper[WrapperActType, ActType, ObsType],
):
    """Base class for environment action wrappers"""

    env: eqx.AbstractVar[AbstractEnvLike[ActType, ObsType]]

    def reset(
        self, state: eqx.nn.State, *, key: Key
    ) -> tuple[eqx.nn.State, ObsType, dict]:
        substate = state.substate(self.env)
        substate, obs, info = self.env.reset(substate, key=key)
        state = state.update(substate)

        return state, obs, info

    def step(
        self, state: eqx.nn.State, action: WrapperActType, *, key: Key
    ) -> tuple[
        eqx.nn.State, ObsType, Float[Array, ""], Bool[Array, ""], Bool[Array, ""], dict
    ]:
        env_key, wrapper_key = jr.split(key, 2)

        state, transformed_action = self.action(state, action, key=wrapper_key)

        env_state = state.substate(self.env)
        env_state, obs, reward, termination, truncation, info = self.env.step(
            env_state, transformed_action, key=env_key
        )
        state = state.update(env_state)

        return state, obs, reward, termination, truncation, info

    @abstractmethod
    def action(
        self, state: eqx.nn.State, action: WrapperActType, *, key: Key
    ) -> tuple[eqx.nn.State, ActType]:
        """Transform the action to the wrapped environment"""


class AbstractPureTransformActionWrapper[WrapperActType, ActType, ObsType](
    AbstractTransformActionWrapper[WrapperActType, ActType, ObsType]
):
    """
    Base class for wrappers that apply a pure function to the action before passing it to
    the environment
    """

    env: eqx.AbstractVar[AbstractEnvLike[ActType, ObsType]]
    func: eqx.AbstractVar[Callable[[WrapperActType], ActType]]
    action_space: eqx.AbstractVar[AbstractSpace[WrapperActType]]

    def action(
        self, state: eqx.nn.State, action: WrapperActType, *, key: Key
    ) -> tuple[eqx.nn.State, ActType]:
        transformed_action = self.func(action)
        return state, transformed_action


class TransformAction[WrapperActType, ActType, ObsType](
    AbstractPureTransformActionWrapper[WrapperActType, ActType, ObsType]
):
    """Apply a function to the action before passing it to the environment"""

    env: AbstractEnvLike[ActType, ObsType]
    func: Callable[[WrapperActType], ActType]
    action_space: AbstractSpace[WrapperActType]

    def __init__(
        self,
        env: AbstractEnvLike[ActType, ObsType],
        func: Callable[[WrapperActType], ActType],
        action_space: AbstractSpace[WrapperActType],
    ):
        self.env = env
        self.func = func
        self.action_space = action_space


class ClipAction[ObsType](
    AbstractPureTransformActionWrapper[
        Float[Array, " ..."], Float[Array, " ..."], ObsType
    ],
):
    """
    Clips every action to the environment's action space.
    """

    env: AbstractEnvLike[Float[Array, " ..."], ObsType]
    func: Callable[[Float[Array, " ..."]], Float[Array, " ..."]]
    action_space: Box

    def __init__(self, env: AbstractEnvLike[Float[Array, " ..."], ObsType]):
        if not isinstance(env.action_space, Box):
            raise ValueError(
                "ClipAction only supports `Box` action spaces "
                f"not {type(env.action_space)}"
            )

        def clip(action: Float[Array, " ..."]) -> Float[Array, " ..."]:
            assert isinstance(env.action_space, Box)
            return jnp.clip(action, env.action_space.low, env.action_space.high)

        action_space = Box(-jnp.inf, jnp.inf, shape=env.action_space.shape)

        self.env = env
        self.func = clip
        self.action_space = action_space


class RescaleAction[ObsType](
    AbstractPureTransformActionWrapper[
        Float[Array, " ..."], Float[Array, " ..."], ObsType
    ],
):
    """Affinely rescale a box action to a different range"""

    env: AbstractEnvLike[Float[Array, " ..."], ObsType]
    func: Callable[[Float[Array, " ..."]], Float[Array, " ..."]]
    action_space: Box

    def __init__(
        self,
        env: AbstractEnvLike[Float[Array, " ..."], ObsType],
        min: Float[Array, " ..."] = jnp.array(-1.0),
        max: Float[Array, " ..."] = jnp.array(1.0),
    ):
        if not isinstance(env.action_space, Box):
            raise ValueError(
                "RescaleActiononly supports `Box` action spaces"
                f" not {type(env.action_space)}"
            )

        action_space, _, rescale = rescale_box(env.action_space, min, max)

        self.env = env
        self.func = rescale
        self.action_space = action_space



from __future__ import annotations

from typing import Callable

import diffrax
import equinox as eqx
import jax
from jax import nn as jnn
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Float, Key

from ..base_model import AbstractModel
from .term import AbstractNODETerm, MLPNODETerm


class AbstractNeuralODE(
    AbstractModel[
        [Float[Array, " n"], Float[Array, " data_size"]], Float[Array, " n data_size"]
    ],
):
    """Neural ODE model"""

    term: eqx.AbstractVar[AbstractNODETerm]
    solver: eqx.AbstractVar[type[diffrax.AbstractSolver]]

    def solve(
        self, ts: Float[Array, " n"], z0: Float[Array, " data_size"]
    ) -> Float[Array, " n data_size"]:
        solver = self.solver()
        if isinstance(solver, diffrax.AbstractAdaptiveSolver):
            stepsize_controller = diffrax.PIDController(rtol=1e-3, atol=1e-6)
            dt0 = None
        else:
            stepsize_controller = diffrax.ConstantStepSize()
            dt0 = jnp.diff(ts).mean()

        term = diffrax.ODETerm(self.term)
        saveat = diffrax.SaveAt(ts=ts)

        solution = diffrax.diffeqsolve(
            term,
            solver,
            t0=jnp.min(ts),
            t1=jnp.max(ts),
            dt0=dt0,
            y0=z0,
            stepsize_controller=stepsize_controller,
            saveat=saveat,
        )

        assert solution.ys is not None

        return solution.ys


class AbstractLatentNeuralODE[
    T: Callable[[Array], Float[Array, " latent_size"]],
    P: Callable[[Array], Float[Array, " latent_size"]],
](AbstractNeuralODE):

    in_size: eqx.AbstractVar[int]
    latent_size: eqx.AbstractVar[int]
    out_size: eqx.AbstractVar[int]

    initial: eqx.AbstractVar[T]
    output: eqx.AbstractVar[P]
    time_in_input: eqx.AbstractVar[bool]

    def __call__(
        self, ts: Float[Array, " n"], x0: Float[Array, " in_size"]
    ) -> Float[Array, " n output_size"]:
        if self.time_in_input:
            z0 = self.initial(
                jnp.concatenate([x0, jnp.expand_dims(ts[0], axis=-1)], axis=-1)
            )
        else:
            z0 = self.initial(x0)

        return jax.vmap(self.output)(self.solve(ts, z0))


class MLPNeuralODE(AbstractLatentNeuralODE[eqx.nn.MLP, eqx.nn.MLP]):
    """MLP based neural ODE."""

    term: MLPNODETerm
    solver: type[diffrax.AbstractSolver]

    in_size: int
    latent_size: int
    out_size: int

    initial: eqx.nn.MLP
    output: eqx.nn.MLP
    time_in_input: bool

    def __init__(
        self,
        in_size: int,
        out_size: int,
        latent_size: int,
        width_size: int,
        depth: int,
        activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = jnn.softplus,
        final_activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = jnn.tanh,
        *,
        key: Key,
        solver: type[diffrax.AbstractSolver] = diffrax.Tsit5,
        time_in_input: bool = True,
    ):
        term_key, initial_key, output_key = jr.split(key, 3)

        self.solver = solver
        self.time_in_input = time_in_input

        self.in_size = in_size
        self.latent_size = latent_size
        self.out_size = out_size

        self.term = MLPNODETerm(
            data_size=latent_size,
            width_size=width_size,
            depth=depth,
            activation=activation,
            final_activation=final_activation,
            key=term_key,
            add_time=time_in_input,
        )
        self.initial = eqx.nn.MLP(
            in_size=in_size + int(time_in_input),
            out_size=latent_size,
            width_size=width_size,
            depth=depth,
            key=initial_key,
        )
        self.output = eqx.nn.MLP(
            in_size=latent_size,
            out_size=out_size,
            width_size=width_size,
            depth=depth,
            key=output_key,
        )



"""
Lerax spaces
"""

from .base_space import AbstractSpace
from .box import Box
from .dict import Dict
from .discrete import Discrete
from .multi_binary import MultiBinary
from .multi_discrete import MultiDiscrete
from .tuple import Tuple

__all__ = [
    "AbstractSpace",
    "Box",
    "Dict",
    "Discrete",
    "MultiBinary",
    "MultiDiscrete",
    "Tuple",
]



from __future__ import annotations

from abc import abstractmethod
from typing import Concatenate

import equinox as eqx


class AbstractModel[**InType, OutType](eqx.Module):
    """Base class for models that take inputs and produce outputs."""

    @abstractmethod
    def __call__(self, *args: InType.args, **kwargs: InType.kwargs) -> OutType:
        """Return an output given an input."""


class AbstractStatefulModel[**InType, *OutType](
    AbstractModel[Concatenate[eqx.nn.State, InType], tuple[eqx.nn.State, *OutType]],
):
    """Base class for models with state."""

    state_index: eqx.AbstractVar[eqx.nn.StateIndex]

    @abstractmethod
    def __call__(
        self, state: eqx.nn.State, *args: InType.args, **kwargs: InType.kwargs
    ) -> tuple[eqx.nn.State, *OutType]:
        """Return an output given inputs and the state."""

    @abstractmethod
    def reset(self, state: eqx.nn.State) -> eqx.nn.State:
        """Reset the state of the model to its initial state."""



from __future__ import annotations

import dataclasses

import equinox as eqx
from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key

from lerax.env import AbstractEnvLike
from lerax.space import AbstractSpace

from .base_wrapper import (
    AbstractNoActionOrObservationSpaceWrapper,
    AbstractNoRenderOrCloseWrapper,
)


class Identity[ActType, ObsType](
    AbstractNoRenderOrCloseWrapper[ActType, ObsType, ActType, ObsType],
    AbstractNoActionOrObservationSpaceWrapper[ActType, ObsType],
):
    env: AbstractEnvLike[ActType, ObsType]

    def __init__(self, env: AbstractEnvLike[ActType, ObsType]):
        self.env = env

    def reset(
        self, state: eqx.nn.State, *, key: Key
    ) -> tuple[eqx.nn.State, ObsType, dict]:
        return self.env.reset(state, key=key)

    def step(
        self, state: eqx.nn.State, action: ActType, *, key: Key
    ) -> tuple[
        eqx.nn.State, ObsType, Float[Array, ""], Bool[Array, ""], Bool[Array, ""], dict
    ]:
        return self.env.step(state, action, key=key)


class LogState(eqx.Module):
    episode_length: Int[Array, ""]
    episode_reward: Float[Array, ""]
    episode_done: Bool[Array, ""]

    def __init__(
        self,
        episode_length: Int[ArrayLike, ""] = 0,
        episode_reward: Float[ArrayLike, ""] = 0.0,
        episode_done: Bool[ArrayLike, ""] = False,
    ):
        # Types must be specified to avoid weak types
        self.episode_length = jnp.array(episode_length, dtype=int)
        self.episode_reward = jnp.array(episode_reward, dtype=float)
        self.episode_done = jnp.array(episode_done, dtype=bool)

    def update(self, reward: Float[ArrayLike, ""], done: Bool[ArrayLike, ""]):
        return dataclasses.replace(
            self,
            episode_length=self.episode_length + 1,
            episode_reward=self.episode_reward + reward,
            episode_done=jnp.array(done, dtype=bool),
        )


class EpisodeStatistics[ActType, ObsType](
    AbstractNoRenderOrCloseWrapper[ActType, ObsType, ActType, ObsType]
):
    state_index: eqx.nn.StateIndex[LogState]
    env: AbstractEnvLike[ActType, ObsType]

    def __init__(self, env: AbstractEnvLike[ActType, ObsType]):
        self.env = env
        self.state_index = eqx.nn.StateIndex(LogState())

    def reset(
        self, state: eqx.nn.State, *, key: Key
    ) -> tuple[eqx.nn.State, ObsType, dict]:
        env_state = state.substate(self.env)
        env_state, obs, info = self.env.reset(env_state, key=key)
        state = state.update(env_state)

        log_state = LogState()
        wrapper_state = state.substate(self.state_index)
        wrapper_state = wrapper_state.set(self.state_index, log_state)
        state = state.update(wrapper_state)

        info["episode"] = {
            "length": log_state.episode_length,
            "reward": log_state.episode_reward,
            "done": log_state.episode_done,
        }

        return state, obs, info

    def step(
        self, state: eqx.nn.State, action: ActType, *, key: Key
    ) -> tuple[
        eqx.nn.State, ObsType, Float[Array, ""], Bool[Array, ""], Bool[Array, ""], dict
    ]:
        env_state = state.substate(self.env)
        env_state, obs, reward, termination, truncation, info = self.env.step(
            env_state, action, key=key
        )
        state = state.update(env_state)

        log_state = state.get(self.state_index)
        log_state = log_state.update(reward, jnp.logical_or(termination, truncation))
        state = state.set(self.state_index, log_state)

        info["episode"] = {
            "length": log_state.episode_length,
            "reward": log_state.episode_reward,
            "done": log_state.episode_done,
        }

        return state, obs, reward, termination, truncation, info

    @property
    def action_space(self) -> AbstractSpace[ActType]:
        return self.env.action_space

    @property
    def observation_space(self) -> AbstractSpace[ObsType]:
        return self.env.observation_space


class TimeLimit[ActType, ObsType](
    AbstractNoActionOrObservationSpaceWrapper[ActType, ObsType],
    AbstractNoRenderOrCloseWrapper[ActType, ObsType, ActType, ObsType],
):

    state_index: eqx.nn.StateIndex[Int[Array, ""]]
    env: AbstractEnvLike[ActType, ObsType]
    max_episode_steps: Int[Array, ""]

    def __init__(self, env: AbstractEnvLike[ActType, ObsType], max_episode_steps: int):
        self.env = env
        self.state_index = eqx.nn.StateIndex(jnp.array(0))
        self.max_episode_steps = jnp.asarray(max_episode_steps)

    def reset(
        self, state: eqx.nn.State, *, key: Key
    ) -> tuple[eqx.nn.State, ObsType, dict]:
        env_state = state.substate(self.env)
        env_state, obs, info = self.env.reset(env_state, key=key)
        state = state.update(env_state)

        state = state.set(self.state_index, jnp.array(0))

        return state, obs, info

    def step(
        self, state: eqx.nn.State, action: ActType, *, key: Key
    ) -> tuple[
        eqx.nn.State, ObsType, Float[Array, ""], Bool[Array, ""], Bool[Array, ""], dict
    ]:
        env_state = state.substate(self.env)
        env_state, obs, reward, termination, truncation, info = self.env.step(
            env_state, action, key=key
        )
        state = state.update(env_state)

        step_count = state.get(self.state_index) + 1
        state = state.set(self.state_index, step_count)

        truncation = jnp.logical_or(truncation, step_count >= self.max_episode_steps)

        return state, obs, reward, termination, truncation, info


class AutoClose[ActType, ObsType](
    AbstractNoActionOrObservationSpaceWrapper[ActType, ObsType],
    AbstractNoRenderOrCloseWrapper[ActType, ObsType, ActType, ObsType],
):
    """
    Closes the environment automatically when it is deleted.
    """

    env: AbstractEnvLike[ActType, ObsType]

    def __init__(self, env: AbstractEnvLike[ActType, ObsType]):
        self.env = env

    def reset(
        self, state: eqx.nn.State, *, key: Key
    ) -> tuple[eqx.nn.State, ObsType, dict]:
        return self.env.reset(state, key=key)

    def step(
        self, state: eqx.nn.State, action: ActType, *, key: Key
    ) -> tuple[
        eqx.nn.State, ObsType, Float[Array, ""], Bool[Array, ""], Bool[Array, ""], dict
    ]:
        return self.env.step(state, action, key=key)

    def close(self):
        self.env.close()

    def __del__(self):
        self.close()



from .base_algorithm import AbstractAlgorithm
from .on_policy import AbstractOnPolicyAlgorithm
from .ppo import PPO

__all__ = [
    "AbstractAlgorithm",
    "AbstractOnPolicyAlgorithm",
    "PPO",
]



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
from jaxtyping import Array, Float, Key

from lerax.distribution import AbstractDistribution
from lerax.space import AbstractSpace

from ..base_policy import AbstractPolicy


# TODO: Add support for split actor and critic observations
class AbstractActorCriticPolicy[FeatureType, ActType, ObsType](
    AbstractPolicy[ActType, ObsType]
):
    """
    Base class for actor-critic policies.

    For now all policies are treated as stateful, meaning they maintain an internal
    state. Non-stateful policies can simply ignore the state during implementation.
    """

    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    @abstractmethod
    def extract_features(
        self, state: eqx.nn.State, observation: ObsType
    ) -> tuple[eqx.nn.State, FeatureType]:
        """Extract features from an observation."""

    @abstractmethod
    def action_dist_from_features(
        self, state: eqx.nn.State, features: FeatureType
    ) -> tuple[
        eqx.nn.State,
        AbstractDistribution[ActType],
    ]:
        """Return an action distribution from features."""

    @abstractmethod
    def value_from_features(
        self, state: eqx.nn.State, features: FeatureType
    ) -> tuple[eqx.nn.State, Float]:
        """Return a value from features."""

    def __call__(
        self, state: eqx.nn.State, observation: ObsType, *, key: Key | None = None
    ) -> tuple[eqx.nn.State, ActType, Float[Array, ""], Float[Array, ""]]:
        """
        Get an action and value from an observation.

        If `key` is provided, it will be used for sampling actions, if no key is
        provided the policy will return the most likely action.
        """
        state, features = self.extract_features(state, observation)
        state, action_dist = self.action_dist_from_features(state, features)
        state, value = self.value_from_features(state, features)

        if key is None:
            action = action_dist.mode()
            log_prob = action_dist.log_prob(action)
        else:
            action, log_prob = action_dist.sample_and_log_prob(key)

        return state, action, value, log_prob.sum().squeeze()

    def predict(
        self, state: eqx.nn.State, observation: ObsType, *, key: Key | None = None
    ) -> tuple[eqx.nn.State, ActType]:
        """Get an action from an observation."""
        state, features = self.extract_features(state, observation)
        state, action_dist = self.action_dist_from_features(state, features)

        if key is None:
            action = action_dist.mode()
        else:
            action = action_dist.sample(key)

        return state, action

    def value(
        self, state: eqx.nn.State, observation: ObsType
    ) -> tuple[eqx.nn.State, Float[Array, ""]]:
        """Get the value of an observation."""
        state, features = self.extract_features(state, observation)
        return self.value_from_features(state, features)

    def evaluate_action(
        self, state: eqx.nn.State, observation: ObsType, action: ActType
    ) -> tuple[eqx.nn.State, Float[Array, ""], Float[Array, ""], Float[Array, ""]]:
        """Evaluate an action given an observation."""
        state, features = self.extract_features(state, observation)
        state, action_dist = self.action_dist_from_features(state, features)
        state, value = self.value_from_features(state, features)
        log_prob = action_dist.log_prob(action)

        try:
            entropy = action_dist.entropy().squeeze()
        except NotImplementedError:
            entropy = -log_prob.mean().squeeze()  # Fallback to negative log prob mean

        return state, value, log_prob.sum().squeeze(), entropy

    @abstractmethod
    def reset(self, state: eqx.nn.State) -> eqx.nn.State:
        """Reset the policy state."""



from .ncde import AbstractNeuralCDE, MLPNeuralCDE
from .term import AbstractNCDETerm, MLPNCDETerm

__all__ = [
    "AbstractNeuralCDE",
    "MLPNeuralCDE",
    "AbstractNCDETerm",
    "MLPNCDETerm",
]



from .actor_critic import (
    AbstractActorCriticPolicy,
    MLPActorCriticPolicy,
)
from .base_policy import AbstractPolicy

__all__ = [
    "AbstractPolicy",
    "AbstractActorCriticPolicy",
    "MLPActorCriticPolicy",
]



"""
Lerax Distributions

Wrapper around distreqx.distributions to allow for easier imports, extended typing, and
future expansion.
"""

from .base_distribution import AbstractDistribution, AbstractTransformedDistribution
from .distributions import (
    Bernoulli,
    Categorical,
    MultivariateNormalDiag,
    Normal,
    SquashedMultivariateNormalDiag,
    SquashedNormal,
)

__all__ = [
    "AbstractDistribution",
    "AbstractTransformedDistribution",
    "Bernoulli",
    "Categorical",
    "Normal",
    "MultivariateNormalDiag",
    "SquashedMultivariateNormalDiag",
    "SquashedNormal",
]



from .base_wrapper import (
    AbstractWrapper,
)
from .misc import (
    AutoClose,
    EpisodeStatistics,
    Identity,
    TimeLimit,
)
from .transform_action import (
    ClipAction,
    RescaleAction,
    TransformAction,
)
from .transform_observation import (
    ClipObservation,
    FlattenObservation,
    RescaleObservation,
)
from .transform_reward import (
    ClipReward,
)

__all__ = [
    "AbstractWrapper",
    "AutoClose",
    "EpisodeStatistics",
    "Identity",
    "TimeLimit",
    "TransformAction",
    "ClipAction",
    "RescaleAction",
    "ClipObservation",
    "FlattenObservation",
    "RescaleObservation",
    "ClipReward",
]



from __future__ import annotations

import equinox as eqx
from distreqx import bijectors, distributions
from jaxtyping import Array, Float, Key


class AbstractDistribution[SampleType](eqx.Module):
    """Base class for all distributions in Lerax."""

    distribution: eqx.AbstractVar[distributions.AbstractDistribution]

    def log_prob(self, value: SampleType) -> Float[Array, ""]:
        """Compute the log probability of a sample."""
        return self.distribution.log_prob(value)

    def prob(self, value: SampleType) -> Float[Array, ""]:
        """Compute the probability of a sample."""
        return self.distribution.prob(value)

    def sample(self, key: Key) -> SampleType:
        """Return a sample from the distribution."""
        return self.distribution.sample(key)

    def entropy(self) -> Float[Array, ""]:
        """Compute the entropy of the distribution."""
        return self.distribution.entropy()

    def mean(self) -> SampleType:
        """Compute the mean of the distribution."""
        return self.distribution.mean()

    def mode(self) -> SampleType:
        """Compute the mode of the distribution."""
        return self.distribution.mode()

    def sample_and_log_prob(self, key: Key) -> tuple[SampleType, Float[Array, ""]]:
        """Return a sample and its log probability."""
        return self.distribution.sample_and_log_prob(key)


class AbstractTransformedDistribution[SampleType](AbstractDistribution[SampleType]):

    distribution: eqx.AbstractVar[distributions.AbstractTransformed]

    def mode(self) -> SampleType:
        # Computing the mode this way is not always correct, but it is a reasonable workaround for the
        # use cases of this library.
        try:
            return super().mode()
        except NotImplementedError:
            # TODO: Add a warning here about the mode not being implemented for
            # the underlying distribution.
            return self.distribution._bijector.forward(
                self.distribution._distribution.mode()
            )

    @property
    def bijector(self) -> bijectors.AbstractBijector:
        return self.distribution.bijector



from .base_env import AbstractEnv, AbstractEnvLike
from .cartpole import CartPole
from .continuous_mountain_car import ContinuousMountainCar
from .mountain_car import MountainCar

__all__ = [
    "AbstractEnvLike",
    "AbstractEnv",
    "CartPole",
    "MountainCar",
    "ContinuousMountainCar",
]



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
from jaxtyping import Key

from lerax.space.base_space import AbstractSpace


# TODO: Break out stateful and non-stateful policies
# TODO: Break out stochastic and non-stochastic policies
class AbstractPolicy[ActType, ObsType](eqx.Module):
    """
    Base class for policies.

    Policies map from observations to actions.
    """

    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    @abstractmethod
    def predict(
        self, state: eqx.nn.State, observation: ObsType, *, key: Key | None = None
    ) -> tuple[eqx.nn.State, ActType]:
        """Choose an action from an observation."""



from __future__ import annotations

from typing import ClassVar, Literal

import equinox as eqx
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Bool, Float, Int, Key

from lerax.render import AbstractRenderer, Color, PygameRenderer, Transform
from lerax.space import Box, Discrete

from .base_env import AbstractEnv


class SOLVER(eqx.Enumeration):
    explicit = "explicit"
    implicit = "implicit"


class CartPole(AbstractEnv[Int[Array, ""], Float[Array, "4"]]):
    name: ClassVar[str] = "CartPole"

    state_index: eqx.nn.StateIndex[Float[Array, "4"]]
    action_space: Discrete
    observation_space: Box

    gravity: float
    masscart: float
    masspole: float
    total_mass: float
    length: float
    polemass_length: float
    force_mag: float
    tau: float
    solver: SOLVER
    theta_threshold_radians: float
    x_threshold: float

    renderer: AbstractRenderer | None

    def __init__(
        self,
        solver: SOLVER = SOLVER.implicit,
        renderer: AbstractRenderer | Literal["auto"] | None = None,
    ):
        self.gravity = 9.8
        self.masscart = 1.0
        self.masspole = 0.1
        self.total_mass = self.masspole + self.masscart
        self.length = 0.5
        self.polemass_length = self.masspole * self.length
        self.force_mag = 10.0
        self.tau = 0.02
        self.solver = solver

        self.theta_threshold_radians = 12 * 2 * jnp.pi / 360
        self.x_threshold = 2.4

        self.action_space = Discrete(2)
        high = jnp.array(
            [
                self.x_threshold * 2,
                jnp.inf,
                self.theta_threshold_radians * 2,
                jnp.inf,
            ],
        )
        self.observation_space = Box(-high, high)

        self.state_index = eqx.nn.StateIndex(jnp.zeros(4))

        if renderer == "auto":
            self.renderer = self.default_renderer()
        else:
            self.renderer = renderer

    def reset(
        self, state: eqx.nn.State, *, key: Key, low: float = -0.05, high: float = 0.05
    ) -> tuple[eqx.nn.State, Float[Array, "4"], dict]:
        new_state = jr.uniform(key, (4,), minval=low, maxval=high)
        state = state.set(self.state_index, new_state)

        return state, new_state, {}

    def step(self, state: eqx.nn.State, action: Int[Array, ""], *, key: Key) -> tuple[
        eqx.nn.State,
        Float[Array, "4"],
        Float[Array, ""],
        Bool[Array, ""],
        Bool[Array, ""],
        dict,
    ]:
        x, x_dot, theta, theta_dot = state.get(self.state_index)

        force = (action * 2 - 1) * self.force_mag
        costheta = jnp.cos(theta)
        sintheta = jnp.sin(theta)

        temp = (
            force + self.polemass_length * jnp.square(theta_dot) * sintheta
        ) / self.total_mass
        thetaacc = (self.gravity * sintheta - costheta * temp) / (
            self.length
            * (4.0 / 3.0 - self.masspole * jnp.square(costheta) / self.total_mass)
        )
        xacc = temp - self.polemass_length * thetaacc * costheta / self.total_mass

        if self.solver == SOLVER.explicit:
            x = x + self.tau * x_dot
            x_dot = x_dot + self.tau * xacc
            theta = theta + self.tau * theta_dot
            theta_dot = theta_dot + self.tau * thetaacc
        else:
            x_dot = x_dot + self.tau * xacc
            x = x + self.tau * x_dot
            theta_dot = theta_dot + self.tau * thetaacc
            theta = theta + self.tau * theta_dot

        state_vals = jnp.asarray([x, x_dot, theta, theta_dot])
        state = state.set(self.state_index, state_vals)

        within_x = jnp.logical_and(x >= -self.x_threshold, x <= self.x_threshold)
        within_theta = jnp.logical_and(
            theta >= -self.theta_threshold_radians,
            theta <= self.theta_threshold_radians,
        )
        terminated = jnp.logical_not(jnp.logical_and(within_x, within_theta))

        reward = jnp.array(1.0)

        return state, state_vals, reward, terminated, jnp.array(False), {}

    def render(self, state: eqx.nn.State):
        x, _, theta, _ = state.get(self.state_index)
        x = x

        assert self.renderer is not None, "Renderer is not set."
        self.renderer.clear()

        # Ground
        self.renderer.draw_line(
            start=jnp.array((-10.0, 0.0)),
            end=jnp.array((10.0, 0.0)),
            color=Color(0.0, 0.0, 0.0),
            width=0.01,
        )
        # Cart
        cart_w, cart_h = 0.3, 0.15
        cart_col = Color(0.0, 0.0, 0.0)
        self.renderer.draw_rect(jnp.array((x, 0.0)), w=cart_w, h=cart_h, color=cart_col)
        # Pole
        pole_start = jnp.asarray((x, cart_h / 4))
        pole_end = pole_start + self.length * jnp.asarray(
            [jnp.sin(theta), jnp.cos(theta)]
        )
        pole_col = Color(0.8, 0.6, 0.4)
        self.renderer.draw_line(pole_start, pole_end, color=pole_col, width=0.05)
        # Pole Hinge
        hinge_r = 0.025
        hinge_col = Color(0.5, 0.5, 0.5)
        self.renderer.draw_circle(pole_start, radius=hinge_r, color=hinge_col)

        self.renderer.render()

    def default_renderer(self) -> AbstractRenderer:
        width, height = 800, 450
        transform = Transform(
            scale=200.0,
            offset=jnp.array([width / 2, height * 0.1]),
            width=width,
            height=height,
            y_up=True,
        )
        return PygameRenderer(
            width=width,
            height=height,
            background_color=Color(1.0, 1.0, 1.0),
            transform=transform,
        )

    def close(self):
        pass



from __future__ import annotations

from typing import Callable

import diffrax
import equinox as eqx
from jax import lax
from jax import nn as jnn
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, Float, Int, Key, PyTree, Shaped

from ..base_model import AbstractStatefulModel
from .term import AbstractNCDETerm, MLPNCDETerm

type Coeffs = tuple[
    PyTree[Shaped[Array, "times-1 ?*shape"]],
    PyTree[Shaped[Array, "times-1 ?*shape"]],
    PyTree[Shaped[Array, "times-1 ?*shape"]],
    PyTree[Shaped[Array, "times-1 ?*shape"]],
]

type CDEState = tuple[
    Float[Array, " n"],  # Time
    Float[Array, " n input_size"],  # Inputs
    Float[Array, " n latent_size"],  # States
]


class AbstractNeuralCDE[
    LatentType: Callable[[Array], Float[Array, " latent_size"]],
    OutputType: Callable[[Array], Float[Array, " latent_size"]],
](
    AbstractStatefulModel[
        [Float[Array, ""], Float[Array, " in_size"]], Float[Array, " out_size"]
    ],
):
    """
    Abstract base class for Neural Controlled Differential Equations (NCDEs).

    x is used to denote the input, t is the time, z is the latent state, and y is the
    output. _s is used to denote all states of a variable (e.g. zs is the latent states
    at all times) and _1 is used to denote the last state of a variable (e.g. z1 is the
    latent state at the last time).

    The state_size defines the maximum number of inputs and corresponding states that
    are tracked. In theory the latest latent vector should contain all the important
    information about the history of inputs and states, but the longer the history
    allows more accurate gradients to be computed for back propagation through time.
    Inference mode can be used to disable the history and only use the latest state for
    faster computation without back propagation through time.
    """

    term: eqx.AbstractVar[AbstractNCDETerm]
    solver: eqx.AbstractVar[type[diffrax.AbstractSolver]]

    in_size: eqx.AbstractVar[int]
    latent_size: eqx.AbstractVar[int]
    out_size: eqx.AbstractVar[int]

    initial: eqx.AbstractVar[LatentType]
    output: eqx.AbstractVar[OutputType]
    time_in_input: eqx.AbstractVar[bool]

    state_size: eqx.AbstractVar[int]
    state_index: eqx.AbstractVar[eqx.nn.StateIndex[CDEState]]

    inference: eqx.AbstractVar[bool]

    def empty_state(self) -> CDEState:
        times = jnp.full((self.state_size,), jnp.nan)
        inputs = jnp.full((self.state_size, self.in_size), jnp.nan)
        states = jnp.full((self.state_size, self.latent_size), jnp.nan)
        return times, inputs, states

    def coeffs(self, ts: Float[Array, " n"], xs: Float[Array, " n in_size"]) -> Coeffs:
        if self.time_in_input:
            xs = jnp.concatenate([ts[:, None], xs], axis=1)

        coeffs = diffrax.backward_hermite_coefficients(ts, xs)

        return coeffs

    def solve(
        self,
        ts: Float[Array, " n"],
        z0: Float[Array, " latent_size"],
        coeffs: Coeffs,
    ) -> Float[Array, " n latent_size"]:
        solver = self.solver()
        if isinstance(solver, diffrax.AbstractAdaptiveSolver):
            stepsize_controller = diffrax.PIDController(rtol=1e-3, atol=1e-6)
            dt0 = None
        else:
            stepsize_controller = diffrax.ConstantStepSize()
            dt0 = jnp.nanmean(jnp.diff(ts))

        control = diffrax.CubicInterpolation(ts, coeffs)
        term = diffrax.ControlTerm(self.term, control).to_ode()
        saveat = diffrax.SaveAt(ts=ts)

        solution = diffrax.diffeqsolve(
            term,
            solver,
            t0=jnp.nanmin(ts),
            t1=jnp.nanmax(ts),
            dt0=dt0,
            y0=z0,
            stepsize_controller=stepsize_controller,
            saveat=saveat,
        )

        assert solution.ys is not None
        zs = solution.ys

        zs = jnp.asarray(jnp.where(jnp.isnan(ts[:, None]), jnp.nan, zs))

        return zs

    def z0(
        self, t0: Float[Array, ""], x0: Float[Array, " in_size"]
    ) -> Float[Array, " latent_size"]:
        if self.time_in_input:
            return self.initial(
                jnp.concatenate([x0, jnp.expand_dims(t0, axis=-1)], axis=-1)
            )
        else:
            return self.initial(x0)

    @staticmethod
    def latest_index(ts: Float[Array, " num_steps"]) -> Int[Array, ""]:
        return jnp.nanargmax(ts)

    def next_state(
        self,
        state: eqx.nn.State,
        t1: Float[Array, ""],
        x1: Float[Array, " in_size"],
    ) -> tuple[
        Float[Array, " num_steps"],
        Float[Array, " num_steps input_size"],
        Float[Array, " num_steps latent_size"],
    ]:
        """Add new time and input pair to the state."""
        ts, xs, zs = state.get(self.state_index)
        latest_index = self.latest_index(ts)
        ts = eqx.error_if(
            ts,
            t1 <= ts[latest_index],
            "new input and time must be later than all previous",
        )

        def shift() -> tuple[
            Float[Array, " num_steps"],
            Float[Array, " num_steps input_size"],
            Float[Array, " num_steps latent_size"],
        ]:
            """Shift the saved times and inputs to make room for the new pair."""
            return (
                jnp.roll(ts, -1).at[-1].set(t1),
                jnp.roll(xs, -1, axis=0).at[-1].set(x1),
                jnp.roll(zs, -1, axis=0).at[-1].set(jnp.nan),
            )

        def insert() -> tuple[
            Float[Array, " num_steps"],
            Float[Array, " num_steps input_size"],
            Float[Array, " num_steps latent_size"],
        ]:
            """Insert the new time and input pair at the end of the saved times and inputs."""
            return ts.at[latest_index + 1].set(t1), xs.at[latest_index + 1].set(x1), zs

        ts, xs, zs = lax.cond(
            latest_index == self.state_size - 1,
            shift,
            insert,
        )

        return ts, xs, zs

    def __call__(
        self,
        state: eqx.nn.State,
        t1: Float[Array, ""],
        x1: Float[Array, " in_size"],
        inference: bool | None = None,
    ) -> tuple[eqx.nn.State, Float[Array, " out_size"]]:
        """
        Compute the next state and output given the current state and input.
        """
        ts, xs, zs = self.next_state(state, t1, x1)

        if inference or self.inference:
            z0 = self.z0(ts[0], xs[0])
            z1 = self.solve(ts, z0, self.coeffs(ts, xs))[self.latest_index(ts)]
        else:
            z0 = zs[self.latest_index(ts)]
            slice_index = jnp.min(
                jnp.asarray([self.state_size - 2, self.latest_index(ts)])
            )
            sliced_ts = lax.dynamic_slice(ts, (slice_index,), (2,))
            sliced_xs = lax.dynamic_slice(xs, (slice_index, 0), (2, self.in_size))

            z1 = self.solve(sliced_ts, z0, self.coeffs(sliced_ts, sliced_xs))[1]

        zs = zs.at[self.latest_index(ts)].set(z1)
        state = state.set(self.state_index, (ts, xs, zs))

        return state, self.output(z1)

    def t1(self, state: eqx.nn.State) -> Float[Array, ""]:
        """Get the last time in the state."""
        ts, _, _ = state.get(self.state_index)
        return jnp.nanmax(ts)

    def ts(self, state: eqx.nn.State) -> Float[Array, " n"]:
        """Get all times in the state."""
        ts, _, _ = state.get(self.state_index)
        return ts

    def x1(self, state: eqx.nn.State) -> Float[Array, " in_size"]:
        """Get the last input in the state."""
        ts, xs, _ = state.get(self.state_index)
        return xs[self.latest_index(ts)]

    def xs(self, state: eqx.nn.State) -> Float[Array, " n in_size"]:
        """Get all inputs in the state."""
        _, xs, _ = state.get(self.state_index)
        return xs

    def z1(
        self, state: eqx.nn.State, inference: bool | None = None
    ) -> Float[Array, " latent_size"]:
        """Get the last state in the state."""
        ts, xs, zs = state.get(self.state_index)

        if inference or self.inference:
            z0 = self.z0(ts[0], xs[0])
            z1 = self.solve(ts, z0, self.coeffs(ts, xs))[self.latest_index(ts)]
        else:
            z1 = zs[self.latest_index(ts)]

        return z1

    def zs(
        self, state: eqx.nn.State, inference: bool | None = None
    ) -> Float[Array, " n latent_size"]:
        """Get all states in the state."""

        ts, xs, zs = state.get(self.state_index)

        if inference or self.inference:
            z0 = self.z0(ts[0], xs[0])
            zs = self.solve(ts, z0, self.coeffs(ts, xs))

        return zs

    def reset(self, state: eqx.nn.State) -> eqx.nn.State:
        """Reset the state to an empty state."""
        return state.set(self.state_index, self.empty_state())


class MLPNeuralCDE(AbstractNeuralCDE):

    term: MLPNCDETerm
    solver: type[diffrax.AbstractSolver]

    initial: eqx.nn.MLP
    output: eqx.nn.MLP
    time_in_input: bool

    in_size: int
    latent_size: int
    out_size: int

    state_size: int
    state_index: eqx.nn.StateIndex[CDEState]

    inference: bool

    def __init__(
        self,
        in_size: int,
        out_size: int,
        latent_size: int,
        width_size: int,
        depth: int,
        field_activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = jnn.softplus,
        field_final_activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = jnn.tanh,
        initial_state_activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = jnn.relu,
        initial_state_final_activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = lambda x: x,
        output_activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = jnn.relu,
        output_final_activation: Callable[
            [Float[Array, " width"]], Float[Array, " width"]
        ] = lambda x: x,
        *,
        solver: type[diffrax.AbstractSolver] = diffrax.Tsit5,
        key: Key,
        time_in_input: bool = False,
        inference: bool = True,
        state_size: int = 16,
    ):
        term_key, initial_key, output_key = jr.split(key, 3)

        self.solver = solver
        self.time_in_input = time_in_input

        self.state_size = state_size
        self.inference = inference

        self.in_size = in_size
        self.latent_size = latent_size
        self.out_size = out_size

        self.term = MLPNCDETerm(
            input_size=in_size,
            data_size=latent_size,
            width_size=width_size,
            depth=depth,
            key=term_key,
            add_time=time_in_input,
            activation=field_activation,
            final_activation=field_final_activation,
        )

        self.initial = eqx.nn.MLP(
            in_size=in_size + int(time_in_input),
            out_size=latent_size,
            width_size=width_size,
            depth=depth,
            key=initial_key,
            activation=initial_state_activation,
            final_activation=initial_state_final_activation,
        )

        self.output = eqx.nn.MLP(
            in_size=latent_size,
            out_size=out_size,
            width_size=width_size,
            depth=depth,
            key=output_key,
            activation=output_activation,
            final_activation=output_final_activation,
        )

        self.state_index = eqx.nn.StateIndex(self.empty_state())



from __future__ import annotations

import equinox as eqx
from jax import numpy as jnp
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Scalar, ScalarLike
from rich import progress, text
from tensorboardX import SummaryWriter

from lerax.utils import (
    debug_with_list_wrapper,
    debug_with_numpy_wrapper,
    debug_wrapper,
)


class EpisodeStatisticsAccumulator(eqx.Module):
    """
    Accumulator for episode statistics.
    This is used to keep track of the episode length, reward, and done status.
    """

    episode_length: Int[Array, ""]
    episode_reward: Float[Array, ""]
    episode_done: Bool[Array, ""]

    @classmethod
    def from_episode_stats(cls, info: dict[str, Array]) -> EpisodeStatisticsAccumulator:
        return cls(
            episode_length=info["length"],
            episode_reward=info["reward"],
            episode_done=info["done"],
        )


class JITSummaryWriter:
    """
    A wrapper around `tensorboardX.SummaryWriter` with a JIT compatible interface.
    """

    summary_writer: SummaryWriter

    def __init__(self, log_dir: str | None = None):
        self.summary_writer = SummaryWriter(log_dir=log_dir)

    def add_scalar(
        self,
        tag: str,
        scalar_value: ScalarLike,
        global_step: Int[ArrayLike, ""] | None = None,
        walltime: Float[ArrayLike, ""] | None = None,
    ):
        """
        Add a scalar value to the summary writer.
        """
        scalar_value = eqx.error_if(
            scalar_value,
            jnp.logical_or(jnp.isnan(scalar_value), jnp.isinf(scalar_value)),
            "Scalar value cannot be NaN",
        )
        debug_with_numpy_wrapper(self.summary_writer.add_scalar, thread=True)(
            tag, scalar_value, global_step, walltime
        )

    def add_dict(
        self,
        scalars: dict[str, Scalar],
        prefix: str = "",
        *,
        global_step: Int[ArrayLike, ""] | None = None,
        walltime: Float[ArrayLike, ""] | None = None,
    ) -> None:
        """
        Log a dictionary of **scalar** values.
        """

        if prefix:
            scalars = {f"{prefix}/{k}": v for k, v in scalars.items()}

        for tag, value in scalars.items():
            self.add_scalar(tag, value, global_step=global_step, walltime=walltime)

    def log_episode_stats(
        self,
        episode_stats: EpisodeStatisticsAccumulator,
        *,
        global_step: Int[ArrayLike, ""] | None = None,
        walltime: Float[ArrayLike, ""] | None = None,
    ) -> None:
        """
        Log episode statistics to the summary writer.
        """

        def log_fn(rewards, lengths, dones, global_step, walltime):
            global_step = None if global_step is None else global_step - len(rewards)
            for reward, length, done in zip(
                rewards,
                lengths,
                dones,
            ):
                global_step = None if global_step is None else global_step + 1
                if done:
                    self.summary_writer.add_scalar(
                        "episode/reward",
                        reward,
                        global_step=global_step,
                        walltime=walltime,
                    )
                    self.summary_writer.add_scalar(
                        "episode/length",
                        length,
                        global_step=global_step,
                        walltime=walltime,
                    )

        debug_with_numpy_wrapper(log_fn, thread=True)(
            episode_stats.episode_reward,
            episode_stats.episode_length,
            episode_stats.episode_done,
            global_step,
            walltime,
        )


def superscript_digit(digit: int) -> str:
    return "⁰¹²³⁴⁵⁶⁷⁸⁹"[digit]


def superscript_int(i: int) -> str:
    return "".join(superscript_digit(int(c)) for c in str(i))


def suffixes(base: int):
    yield ""

    val = 1
    while True:
        yield f"×{base}{superscript_int(val)}"
        val += 1


def unit_and_suffix(value: float, base: int) -> tuple[float, str]:
    if base < 1:
        raise ValueError("base must be >= 1")

    unit, suffix = 1, ""
    for i, suffix in enumerate(suffixes(base)):
        unit = base**i
        if int(value) < unit * base:
            break

    return unit, suffix


class SpeedColumn(progress.ProgressColumn):
    """
    Renders human readable speed.

    https://github.com/NichtJens/rich/tree/master
    """

    def render(self, task: progress.Task) -> text.Text:
        """Show speed."""
        speed = task.finished_speed or task.speed

        if speed is None:
            return text.Text("", style="progress.percentage")
        unit, suffix = unit_and_suffix(speed, 2)
        data_speed = speed / unit
        return text.Text(f"{data_speed:.1f}{suffix} it/s", style="red")


class JITProgressBar:
    progress_bar: progress.Progress
    task: progress.TaskID

    def __init__(self, name: str, total: int | None, transient: bool = False):
        self.progress_bar = progress.Progress(
            progress.TextColumn("[progress.description]{task.description}"),
            progress.SpinnerColumn(finished_text="[green]✔"),
            progress.MofNCompleteColumn(),
            progress.BarColumn(bar_width=None),
            progress.TaskProgressColumn(),
            progress.TextColumn("["),
            progress.TimeElapsedColumn(),
            progress.TextColumn("<"),
            progress.TimeRemainingColumn(),
            progress.TextColumn("]"),
            SpeedColumn(),
            transient=transient,
        )
        self.task = self.progress_bar.add_task(f"[yellow]{name}", total=total)

    def start(self) -> None:
        debug_wrapper(self.progress_bar.start, thread=True)()

    def stop(self) -> None:
        debug_wrapper(self.progress_bar.stop, thread=True)()

    def update(
        self,
        total: Float[ArrayLike, ""] | None = None,
        completed: Float[ArrayLike, ""] | None = None,
        advance: Float[ArrayLike, ""] | None = None,
        description: str | None = None,
        visible: Bool[ArrayLike, ""] | None = None,
        refresh: Bool[ArrayLike, ""] = False,
    ) -> None:
        debug_with_list_wrapper(self.progress_bar.update, thread=True)(
            self.task,
            total=total,
            completed=completed,
            advance=advance,
            description=description,
            visible=visible,
            refresh=refresh,
        )



from __future__ import annotations

import dataclasses
from functools import partial

import equinox as eqx
import jax
from jax import lax
from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Key, PyTree

from lerax.utils import clone_state

from .base_buffer import AbstractBuffer


class RolloutBuffer[ActType, ObsType](AbstractBuffer):
    """
    RolloutBuffer used by on-policy algorithms.

    Designed for scans and JIT compilation.
    """

    observations: PyTree[ObsType]
    actions: PyTree[ActType]
    rewards: Float[Array, " *size"]
    terminations: Bool[Array, " *size"]
    truncations: Bool[Array, " *size"]
    log_probs: Float[Array, " *size"]
    values: Float[Array, " *size"]
    returns: Float[Array, " *size"]
    advantages: Float[Array, " *size"]
    states: PyTree[eqx.nn.State]

    def __init__(
        self,
        observations: PyTree[ObsType],
        actions: PyTree[ActType],
        rewards: Float[ArrayLike, " *size"],
        terminations: Bool[ArrayLike, " *size"],
        truncations: Bool[ArrayLike, " *size"],
        log_probs: Float[ArrayLike, " *size"],
        values: Float[ArrayLike, " *size"],
        states: PyTree[eqx.nn.State],
        returns: Float[ArrayLike, " *size"] | None = None,
        advantages: Float[ArrayLike, " *size"] | None = None,
    ):
        """
        Initialize the RolloutBuffer with the given parameters.

        Returns and advantages can be provided, but if not, they will be filled with
        NaNs.
        """
        # TODO: Add type checks for observations, actions, etc
        # TODO: Add shape checks for rewards, terminations, truncations, log_probs,
        # values

        self.observations = observations
        self.actions = actions
        self.rewards = jnp.asarray(rewards)
        self.terminations = jnp.asarray(terminations)
        self.truncations = jnp.asarray(truncations)
        self.log_probs = jnp.asarray(log_probs)
        self.values = jnp.asarray(values)
        # Assume the state has been used already and clone to compensate
        self.states = clone_state(states)
        self.returns = (
            jnp.asarray(returns)
            if returns is not None
            else jnp.full_like(values, jnp.nan)
        )
        self.advantages = (
            jnp.asarray(advantages)
            if advantages is not None
            else jnp.full_like(values, jnp.nan)
        )

    def compute_returns_and_advantages(
        self,
        last_value: Float[ArrayLike, ""],
        done: Bool[ArrayLike, ""],
        gae_lambda: Float[ArrayLike, ""],
        gamma: Float[ArrayLike, ""],
    ) -> RolloutBuffer[ActType, ObsType]:
        """
        Compute returns and advantages for the rollout buffer using Generalized
        Advantage Estimation.

        Works under JIT compilation.
        """
        last_value = jnp.asarray(last_value)
        done = jnp.asarray(done)
        gamma = jnp.asarray(gamma)
        gae_lambda = jnp.asarray(gae_lambda)

        dones = jnp.logical_or(self.terminations, self.truncations)

        next_values = jnp.concatenate(
            [self.values[1:], jnp.array([last_value])], axis=0
        )

        next_non_terminal = jnp.concatenate(
            [1.0 - dones[1:], jnp.array([1.0 - done])], axis=0
        )

        deltas = self.rewards + gamma * next_values * next_non_terminal - self.values

        def scan_fn(
            advantage_carry: Float[Array, ""],
            x: tuple[Float[Array, ""], Float[Array, ""]],
        ) -> tuple[Float[Array, ""], Float[Array, ""]]:
            delta, next_non_terminal = x
            advantage = delta + gamma * gae_lambda * next_non_terminal * advantage_carry
            return advantage, advantage

        _, advantages = lax.scan(
            scan_fn, jnp.array(0.0), (jnp.flip(deltas), jnp.flip(next_non_terminal))
        )
        advantages = jnp.flip(advantages)
        returns = advantages + self.values

        return dataclasses.replace(self, advantages=advantages, returns=returns)

    def batches(
        self, batch_size: int, *, key: Key | None = None
    ) -> RolloutBuffer[ActType, ObsType]:
        """
        Return rollout buffer with batches of the given size.

        The buffer is shuffled if a key is provided.

        This method reshapes the buffer into batches of the specified size. It is not
        the same as a list of butches, but rather a single buffer where each batch is a
        slice of the original data.
        """
        if key is None:
            indices = jnp.arange(self.shape[0])
        else:
            indices = jr.permutation(key, self.shape[0])

        if self.shape[0] % batch_size != 0:
            # TODO: Add warning if batch_size does not divide the number of samples
            indices = indices[: self.shape[0] - self.shape[0] % batch_size]

        indices = indices.reshape(-1, batch_size)

        return jax.tree.map(partial(jnp.take, indices=indices, axis=0), self)

    @property
    def shape(self) -> tuple[int, ...]:
        return self.rewards.shape



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
from jaxtyping import Array, Bool, Float, Key

from lerax.space import AbstractSpace


class AbstractEnvLike[ActType, ObsType](eqx.Module):
    """Base class for RL environments or wrappers that behave like environments"""

    name: eqx.AbstractVar[str]

    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    @abstractmethod
    def reset(
        self, state: eqx.nn.State, *, key: Key
    ) -> tuple[eqx.nn.State, ObsType, dict]:
        """Reset the environment to an initial state"""

    @abstractmethod
    def step(
        self, state: eqx.nn.State, action: ActType, *, key: Key
    ) -> tuple[
        eqx.nn.State, ObsType, Float[Array, ""], Bool[Array, ""], Bool[Array, ""], dict
    ]:
        """
        Perform a step of the environment
        """

    @abstractmethod
    def render(self, state: eqx.nn.State):
        """Render a frame from a state"""

    @abstractmethod
    def close(self):
        """Close the environment"""

    @property
    @abstractmethod
    def unwrapped(self) -> AbstractEnv[ActType, ObsType]:
        """Return the unwrapped environment"""


class AbstractEnv[ActType, ObsType](AbstractEnvLike[ActType, ObsType]):
    """Base class for RL environments"""

    name: eqx.AbstractVar[str]

    action_space: eqx.AbstractVar[AbstractSpace[ActType]]
    observation_space: eqx.AbstractVar[AbstractSpace[ObsType]]

    @property
    def unwrapped(self) -> AbstractEnv[ActType, ObsType]:
        """Return the unwrapped environment"""
        return self



from __future__ import annotations

from jax import numpy as jnp
from jaxtyping import Array, Float

from .base_model import AbstractModel


class Flatten(AbstractModel[[Float[Array, "..."]], Float[Array, " out_size"]]):
    """Warps the JAX `jnp.ravel` function to flatten an input array."""

    def __call__(self, x: Float[Array, "..."]) -> Float[Array, " out_size"]:
        """Flatten the input array into a 1D array."""
        return jnp.ravel(x)



from __future__ import annotations

from typing import Any

from jax import numpy as jnp
from jax import random as jr
from jaxtyping import Array, ArrayLike, Bool, Float, Int, Key

from .base_space import AbstractSpace


class Dict(AbstractSpace[dict[str, Any]]):
    """A dictionary of spaces."""

    spaces: dict[str, AbstractSpace]

    def __init__(self, spaces: dict[str, AbstractSpace]):
        assert isinstance(spaces, dict), "spaces must be a dict"
        assert len(spaces) > 0, "spaces must be non-empty"
        assert all(
            isinstance(space, AbstractSpace) for space in spaces.values()
        ), "spaces must be a dict of AbstractSpace"

        self.spaces = spaces

    @property
    def shape(self) -> None:
        return None

    def canonical(self) -> dict[str, Any]:
        return {key: space.canonical() for key, space in self.spaces.items()}

    def sample(self, key: Key) -> dict[str, Any]:
        return {
            space_key: self.spaces[space_key].sample(rng_key)
            for space_key, rng_key in zip(
                self.spaces.keys(), jr.split(key, len(self.spaces))
            )
        }

    def contains(self, x: Any) -> Bool[ArrayLike, ""]:
        if not isinstance(x, dict):
            return False

        if len(x) != len(self.spaces):
            return False

        return jnp.array(
            key in self.spaces and self.spaces[key].contains(x[key]) for key in x.keys()
        ).all()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Dict):
            return False

        return all(
            key in other.spaces and self.spaces[key] == other.spaces[key]
            for key in self.spaces.keys()
        )

    def __repr__(self) -> str:
        return f"Dict({', '.join(f'{key}: {repr(space)}' for key, space in self.spaces.items())})"

    def __hash__(self) -> int:
        return hash(tuple((key, hash(space)) for key, space in self.spaces.items()))

    def flatten_sample(self, sample: dict[str, Any]) -> Float[Array, " size"]:
        parts = [
            subspace.flatten_sample(sample[key])
            for key, subspace in sorted(self.spaces.items())
        ]
        return jnp.concatenate(parts)

    @property
    def flat_size(self) -> Int[ArrayLike, ""]:
        return (
            jnp.array(space.flat_size for space in self.spaces.values())
            .sum()
            .astype(int)
        )

    def __getitem__(self, index: str) -> AbstractSpace:
        return self.spaces[index]

    def __len__(self) -> int:
        return len(self.spaces)
